r"""
Groebner Fans

Sage provides much of the functionality of ``gfan``, which is a
software package whose main function is to enumerate all reduced
Groebner bases of a polynomial ideal. The reduced Groebner bases
yield the maximal cones in the Groebner fan of the ideal. Several
subcomputations can be issued and additional tools are included.
Among these the highlights are:

-  Commands for computing tropical varieties.

-  Interactive walks in the Groebner fan of an ideal.

-  Commands for graphical renderings of Groebner fans and monomial
   ideals.


AUTHORS:

- Anders Nedergaard Jensen: Wrote the ``gfan`` C++ program, which
  implements algorithms many of which were invented by Jensen, Komei
  Fukuda, and Rekha Thomas. All the underlying hard work of the
  Groebner fans functionality of Sage depends on this C++ program.

- William Stein (2006-04-20): Wrote first version of the Sage code
  for working with Groebner fans.

- Tristram Bogart: the design of the Sage interface
  to ``gfan`` is joint work with Tristram Bogart, who also supplied
  numerous examples.

- Marshall Hampton (2008-03-25): Rewrote various functions to use
  ``gfan-0.3``. This is still a work in progress, comments are
  appreciated on sage-devel@googlegroups.com (or personally at
  hamptonio@gmail.com).

EXAMPLES::

    sage: x,y = QQ['x,y'].gens()
    sage: i = ideal(x^2 - y^2 + 1)
    sage: g = i.groebner_fan()
    sage: g.reduced_groebner_bases()
    [[x^2 - y^2 + 1], [-x^2 + y^2 - 1]]

TESTS::

    sage: x,y = QQ['x,y'].gens()
    sage: i = ideal(x^2 - y^2 + 1)
    sage: g = i.groebner_fan()
    sage: g == loads(dumps(g))
    True

REFERENCES:

- Anders N. Jensen; *Gfan, a software system for Groebner fans*;
  http://home.math.au.dk/jensen/software/gfan/gfan.html
"""
from subprocess import PIPE, Popen
import pexpect
import re
import string
from collections.abc import Iterator

from sage.structure.sage_object import SageObject
from sage.interfaces.gfan import gfan
from .multi_polynomial_ideal import MPolynomialIdeal
from .polynomial_ring_constructor import PolynomialRing
from sage.rings.rational_field import QQ
from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ
from sage.modules.free_module_element import vector
from sage.misc.lazy_import import lazy_import
lazy_import("sage.plot.all", ["line", "Graphics", "polygon"])
lazy_import("sage.plot.plot3d.shapes2", "line3d")
from sage.geometry.polyhedron.constructor import Polyhedron
from sage.geometry.fan import Fan
from sage.matrix.constructor import matrix
from sage.misc.cachefunc import cached_method
from sage.misc.misc_c import prod


def prefix_check(str_list) -> bool:
    """
    Check if any strings in a list are prefixes of another string in
    the list.

    EXAMPLES::

        sage: from sage.rings.polynomial.groebner_fan import prefix_check
        sage: prefix_check(['z1','z1z1'])
        False
        sage: prefix_check(['z1','zz1'])
        True
    """
    for index1, string1 in enumerate(str_list):
        for index2, string2 in enumerate(str_list):
            if index1 != index2:
                if string1[:len(string2)] == string2:
                    return False
    return True


def max_degree(list_of_polys) -> float:
    """
    Compute the maximum degree of a list of polynomials.

    EXAMPLES::

        sage: from sage.rings.polynomial.groebner_fan import max_degree
        sage: R.<x,y> = PolynomialRing(QQ,2)
        sage: p_list = [x^2-y,x*y^10-x]
        sage: max_degree(p_list)
        11.0
    """
    return float(max(qf.degree() for qf in list_of_polys))


def _cone_parse(fan_dict_cone) -> dict:
    """
    Utility function that parses cone information into a dict indexed by
    dimension.

    INPUT:

    - ``fan_dict_cone`` -- the value of a ``fan_dict`` with key 'CONES'

    EXAMPLES::

        sage: R.<x,y,z,w> = PolynomialRing(QQ,4)
        sage: I = R.ideal([x^2+y^2+z^2-1,x^4-y^2-z-1,x+y+z,w+x+y])
        sage: GF = I.groebner_fan()
        sage: TI = GF.tropical_intersection()
        sage: from sage.rings.polynomial.groebner_fan import _cone_parse
        sage: _cone_parse(TI.fan_dict['CONES'])
        {1: [[0], [1], [2], [3], [4]], 2: [[2, 3]]}
    """
    cone_dict = {}
    cur_dim = 0
    for item in fan_dict_cone:
        temp = item.split('{')[-1]
        if temp.split('}') == '':
            temp = ['']
        else:
            temp = temp.split('}')[0]
            temp = temp.split(' ')
        if item.find('Dimension') != -1:
            cur_dim = Integer(item.split(' ')[-1])
            if cur_dim > 0:
                cone_dict[cur_dim] = [[Integer(q) for q in temp if q != '']]
        else:
            if cur_dim > 0:
                cone_dict[cur_dim] += [[Integer(q) for q in temp if q != '']]
    return cone_dict


class PolyhedralCone(SageObject):

    def __init__(self, gfan_polyhedral_cone, ring=QQ) -> None:
        """
        Convert polymake/gfan data on a polyhedral cone into a sage class.

        Currently (18-03-2008) needs a lot of work.

        EXAMPLES::

            sage: R3.<x,y,z> = PolynomialRing(QQ,3)
            sage: gf = R3.ideal([x^8-y^4,y^4-z^2,z^2-2]).groebner_fan()
            sage: a = gf[0].groebner_cone()
            sage: a.facets()
            [[0, 0, 1], [0, 1, 0], [1, 0, 0]]
        """
        cone_keys = ['AMBIENT_DIM', 'DIM', 'IMPLIED_EQUATIONS',
                     'LINEALITY_DIM', 'LINEALITY_SPACE', 'FACETS',
                     'RELATIVE_INTERIOR_POINT']
        poly_lines = gfan_polyhedral_cone.split('\n')
        self.cone_dict = {}
        cur_key = None
        for ting in poly_lines:
            if cone_keys.count(ting):
                cur_key = ting
                self.cone_dict[cur_key] = []
            elif cur_key and ting:
                self.cone_dict[cur_key].append(ting)
        self._facets = []
        for facet in self.cone_dict['FACETS']:
            temp_facet = facet.split('\t')[0]
            temp_facet = temp_facet.split(' ')
            temp_facet = [int(x) for x in temp_facet]
            self._facets.append(temp_facet)
        self._ambient_dim = int(self.cone_dict['AMBIENT_DIM'][0])
        self._dim = int(self.cone_dict['DIM'][0])
        self._lineality_dim = int(self.cone_dict['LINEALITY_DIM'][0])
        rel_int_pt_str = self.cone_dict['RELATIVE_INTERIOR_POINT'][0]
        self._relative_interior_point = [int(q) for q in rel_int_pt_str.split(' ')]

    def _repr_(self) -> str:
        """
        Return a basic description of the polyhedral cone.

        EXAMPLES::

            sage: R3.<x,y,z> = PolynomialRing(QQ,3)
            sage: gf = R3.ideal([x^8-y^4,y^4-z^2,z^2-2]).groebner_fan()
            sage: a = gf[0].groebner_cone()
            sage: a # indirect doctests
            Polyhedral cone in 3 dimensions of dimension 3
        """
        return "Polyhedral cone in {} dimensions of dimension {}".format(self.ambient_dim(), self.dim())

    def facets(self) -> list:
        """
        Return the inward facet normals of the Groebner cone.

        EXAMPLES::

            sage: R3.<x,y,z> = PolynomialRing(QQ,3)
            sage: gf = R3.ideal([x^8-y^4,y^4-z^2,z^2-2]).groebner_fan()
            sage: a = gf[0].groebner_cone()
            sage: a.facets()
            [[0, 0, 1], [0, 1, 0], [1, 0, 0]]
        """
        return self._facets

    def ambient_dim(self):
        """
        Return the ambient dimension of the Groebner cone.

        EXAMPLES::

            sage: R3.<x,y,z> = PolynomialRing(QQ,3)
            sage: gf = R3.ideal([x^8-y^4,y^4-z^2,z^2-2]).groebner_fan()
            sage: a = gf[0].groebner_cone()
            sage: a.ambient_dim()
            3
        """
        return self._ambient_dim

    def dim(self):
        """
        Return the dimension of the Groebner cone.

        EXAMPLES::

            sage: R3.<x,y,z> = PolynomialRing(QQ,3)
            sage: gf = R3.ideal([x^8-y^4,y^4-z^2,z^2-2]).groebner_fan()
            sage: a = gf[0].groebner_cone()
            sage: a.dim()
            3
        """
        return self._dim

    def lineality_dim(self):
        """
        Return the lineality dimension of the Groebner cone. This is
        just the difference between the ambient dimension and the dimension
        of the cone.

        EXAMPLES::

            sage: R3.<x,y,z> = PolynomialRing(QQ,3)
            sage: gf = R3.ideal([x^8-y^4,y^4-z^2,z^2-2]).groebner_fan()
            sage: a = gf[0].groebner_cone()
            sage: a.lineality_dim()
            0
        """
        return self._lineality_dim

    def relative_interior_point(self):
        """
        Return a point in the relative interior of the Groebner cone.

        EXAMPLES::

            sage: R3.<x,y,z> = PolynomialRing(QQ,3)
            sage: gf = R3.ideal([x^8-y^4,y^4-z^2,z^2-2]).groebner_fan()
            sage: a = gf[0].groebner_cone()
            sage: a.relative_interior_point()
            [1, 1, 1]
        """
        return self._relative_interior_point


class PolyhedralFan(SageObject):
    def __init__(self, gfan_polyhedral_fan, parameter_indices=None) -> None:
        """
        Convert polymake/gfan data on a polyhedral fan into a sage class.

        INPUT:

        - ``gfan_polyhedral_fan`` -- output from gfan of a polyhedral fan

        EXAMPLES::

            sage: R.<x,y,z> = PolynomialRing(QQ,3)
            sage: i2 = ideal(x*z + 6*y*z - z^2, x*y + 6*x*z + y*z - z^2, y^2 + x*z + y*z)
            sage: gf2 = i2.groebner_fan(verbose=False)
            sage: pf = gf2.polyhedralfan()
            sage: pf.rays()
            [[-1, 0, 1], [-1, 1, 0], [1, -2, 1], [1, 1, -2], [2, -1, -1]]
        """
        if parameter_indices is None:
            parameter_indices = []
        fan_keys = ['AMBIENT_DIM', 'DIM', 'LINEALITY_DIM', 'RAYS', 'N_RAYS',
                    'LINEALITY_SPACE', 'ORTH_LINEALITY_SPACE', 'F_VECTOR',
                    'CONES', 'MAXIMAL_CONES', 'PURE', 'SIMPLICIAL',
                    'MULTIPLICITIES']
        poly_lines = gfan_polyhedral_fan.split('\n')
        self.fan_dict = {}
        cur_key = None
        for ting in poly_lines:
            if fan_keys.count(ting):
                cur_key = ting
                self.fan_dict[cur_key] = []
            elif cur_key and ting != '':
                self.fan_dict[cur_key].append(ting)
        self._ambient_dim = int(self.fan_dict['AMBIENT_DIM'][0])
        self._dim = int(self.fan_dict['DIM'][0])
        self._lineality_dim = int(self.fan_dict['LINEALITY_DIM'][0])
        self._rays = []
        for ray in self.fan_dict['RAYS']:
            temp_ray = ray.split('\t')[0]
            temp_ray = temp_ray.split(' ')
            temp_ray = [int(x) for x in temp_ray]
            if parameter_indices != []:
                for q in parameter_indices:
                    temp_ray = temp_ray[0:q] + [0] + temp_ray[q:]
            self._rays.append(temp_ray)
        self._cone_dict = _cone_parse(self.fan_dict['CONES'])
        self._maximal_cone_dict = _cone_parse(self.fan_dict['MAXIMAL_CONES'])
        self._str = gfan_polyhedral_fan

    def _repr_(self) -> str:
        """
        Return a basic description of the polyhedral fan.

        EXAMPLES::

            sage: R3.<x,y,z> = PolynomialRing(QQ,3)
            sage: gf = R3.ideal([x^8-y^4,y^4-z^2,z^2-2]).groebner_fan()
            sage: pf = gf.polyhedralfan()
            sage: pf # indirect doctest
            Polyhedral fan in 3 dimensions of dimension 3
        """
        return "Polyhedral fan in {} dimensions of dimension {}".format(self.ambient_dim(), self.dim())

    def _str_(self) -> str:
        r"""
        Return the raw output of gfan as a string.

        This should only be needed internally as all relevant output
        is converted to sage objects.

        EXAMPLES::

            sage: R3.<x,y,z> = PolynomialRing(QQ,3)
            sage: gf = R3.ideal([x^8-y^4,y^4-z^2,z^2-2]).groebner_fan()
            sage: pf = gf.polyhedralfan()
            sage: pf._str_()
            '_application fan\n_version 2.2\n_type SymmetricFan\n\nAMBIENT_DIM\n3\n\nDIM\n3\n\nLINEALITY_DIM\n0\n\nRAYS\n0 0 1\t# 0\n0 1 0\t# 1\n1 0 0\t# 2\n\nN_RAYS\n3\n\nLINEALITY_SPACE\n\nORTH_LINEALITY_SPACE\n1 0 0\n0 1 0\n0 0 1\n\nF_VECTOR\n1 3 3 1\n\nSIMPLICIAL\n1\n\nPURE\n1\n\nCONES\n{}\t# Dimension 0\n{0}\t# Dimension 1\n{1}\n{2}\n{0 1}\t# Dimension 2\n{0 2}\n{1 2}\n{0 1 2}\t# Dimension 3\n\nMAXIMAL_CONES\n{0 1 2}\t# Dimension 3\n'
        """
        return self._str

    def ambient_dim(self):
        """
        Return the ambient dimension of the Groebner fan.

        EXAMPLES::

            sage: R3.<x,y,z> = PolynomialRing(QQ,3)
            sage: gf = R3.ideal([x^8-y^4,y^4-z^2,z^2-2]).groebner_fan()
            sage: a = gf.polyhedralfan()
            sage: a.ambient_dim()
            3
        """
        return self._ambient_dim

    def dim(self):
        """
        Return the dimension of the Groebner fan.

        EXAMPLES::

            sage: R3.<x,y,z> = PolynomialRing(QQ,3)
            sage: gf = R3.ideal([x^8-y^4,y^4-z^2,z^2-2]).groebner_fan()
            sage: a = gf.polyhedralfan()
            sage: a.dim()
            3
        """
        return self._dim

    def lineality_dim(self):
        """
        Return the lineality dimension of the fan. This is the
        dimension of the largest subspace contained in the fan.

        EXAMPLES::

            sage: R3.<x,y,z> = PolynomialRing(QQ,3)
            sage: gf = R3.ideal([x^8-y^4,y^4-z^2,z^2-2]).groebner_fan()
            sage: a = gf.polyhedralfan()
            sage: a.lineality_dim()
            0
        """
        return self._lineality_dim

    def rays(self) -> list:
        """
        A list of rays of the polyhedral fan.

        EXAMPLES::

            sage: R.<x,y,z> = PolynomialRing(QQ,3)
            sage: i2 = ideal(x*z + 6*y*z - z^2, x*y + 6*x*z + y*z - z^2, y^2 + x*z + y*z)
            sage: gf2 = i2.groebner_fan(verbose=False)
            sage: pf = gf2.polyhedralfan()
            sage: pf.rays()
            [[-1, 0, 1], [-1, 1, 0], [1, -2, 1], [1, 1, -2], [2, -1, -1]]
        """
        return sorted(self._rays)

    def cones(self) -> dict:
        """
        A dictionary of cones in which the keys are the cone dimensions.

        For each dimension, the value is a list of the cones,
        where each element consists of a list of ray indices.

        EXAMPLES::

            sage: R.<x,y,z> = QQ[]
            sage: f = 1+x+y+x*y
            sage: I = R.ideal([f+z*f, 2*f+z*f, 3*f+z^2*f])
            sage: GF = I.groebner_fan()
            sage: PF = GF.tropical_intersection()
            sage: PF.cones()
            {1: [[0], [1], [2], [3], [4], [5]], 2: [[0, 1], [0, 2], [0, 3], [0, 4], [1, 2], [1, 3], [2, 4], [3, 4], [1, 5], [2, 5], [3, 5], [4, 5]]}
        """
        return self._cone_dict

    def maximal_cones(self) -> dict:
        """
        A dictionary of the maximal cones in which the keys are the
        cone dimensions.

        For each dimension, the value is a list of
        the maximal cones, where each element consists of a list of ray indices.

        EXAMPLES::

            sage: R.<x,y,z> = QQ[]
            sage: f = 1+x+y+x*y
            sage: I = R.ideal([f+z*f, 2*f+z*f, 3*f+z^2*f])
            sage: GF = I.groebner_fan()
            sage: PF = GF.tropical_intersection()
            sage: PF.maximal_cones()
            {2: [[0, 1], [0, 2], [0, 3], [0, 4], [1, 2], [1, 3], [2, 4], [3, 4], [1, 5], [2, 5], [3, 5], [4, 5]]}
        """
        return self._maximal_cone_dict

    def f_vector(self) -> list:
        """
        The f-vector of the fan.

        EXAMPLES::

            sage: R.<x,y,z> = QQ[]
            sage: f = 1+x+y+x*y
            sage: I = R.ideal([f+z*f, 2*f+z*f, 3*f+z^2*f])
            sage: GF = I.groebner_fan()
            sage: PF = GF.tropical_intersection()
            sage: PF.f_vector()
            [1, 6, 12]
        """
        str_data = self.fan_dict['F_VECTOR'][0]
        return [Integer(x) for x in str_data.split(' ')]

    def is_simplicial(self) -> bool:
        """
        Whether the fan is simplicial or not.

        EXAMPLES::

            sage: R.<x,y,z> = QQ[]
            sage: f = 1+x+y+x*y
            sage: I = R.ideal([f+z*f, 2*f+z*f, 3*f+z^2*f])
            sage: GF = I.groebner_fan()
            sage: PF = GF.tropical_intersection()
            sage: PF.is_simplicial()
            True
        """
        return bool(int(self.fan_dict['SIMPLICIAL'][0]))

    def to_RationalPolyhedralFan(self):
        """
        Convert to the RationalPolyhedralFan class, which is more actively
        maintained.

        While the information in each class is essentially the
        same, the methods and implementation are different.

        EXAMPLES::

            sage: R.<x,y,z> = QQ[]
            sage: f = 1+x+y+x*y
            sage: I = R.ideal([f+z*f, 2*f+z*f, 3*f+z^2*f])
            sage: GF = I.groebner_fan()
            sage: PF = GF.tropical_intersection()
            sage: fan = PF.to_RationalPolyhedralFan()
            sage: [tuple(q.facet_normals()) for q in fan]
            [(M(0, -1, 0), M(-1, 0, 0)), (M(0, 0, -1), M(-1, 0, 0)), (M(0, 0, 1), M(-1, 0, 0)), (M(0, 1, 0), M(-1, 0, 0)), (M(0, 0, -1), M(0, -1, 0)), (M(0, 0, 1), M(0, -1, 0)), (M(0, 1, 0), M(0, 0, -1)), (M(0, 1, 0), M(0, 0, 1)), (M(1, 0, 0), M(0, -1, 0)), (M(1, 0, 0), M(0, 0, -1)), (M(1, 0, 0), M(0, 0, 1)), (M(1, 0, 0), M(0, 1, 0))]

        Here we use the RationalPolyhedralFan's Gale_transform method on a tropical
        prevariety.

        .. link

        ::

            sage: fan.Gale_transform()
            [ 1  0  0  0  0  1 -2]
            [ 0  1  0  0  1  0 -2]
            [ 0  0  1  1  0  0 -2]
        """
        try:
            return self._fan
        except AttributeError:
            cdnt = []
            cones = self.cones()
            for x in cones:
                if x > 1:
                    cdnt += cones[x]
            fan = Fan(cones=cdnt, rays=self.rays(), discard_faces=True)
            self._fan = fan
            return self._fan


class InitialForm(SageObject):
    def __init__(self, cone, rays, initial_forms) -> None:
        """
        A system of initial forms from a polynomial system.

        To each form is associated a cone and a list of
        polynomials (the initial form system itself).

        This class is intended for internal use inside of the
        :class:`TropicalPrevariety` class.

        EXAMPLES::

            sage: from sage.rings.polynomial.groebner_fan import InitialForm
            sage: R.<x,y> = QQ[]
            sage: inform = InitialForm([0], [[-1, 0]], [y^2 - 1, y^2 - 2, y^2 - 3])
            sage: inform._cone
            [0]
        """
        self._cone = cone
        self._rays = rays
        self._initial_forms = initial_forms

    def cone(self):
        """
        The cone associated with the initial form system.

        EXAMPLES::

            sage: R.<x,y> = QQ[]
            sage: I = R.ideal([(x+y)^2-1,(x+y)^2-2,(x+y)^2-3])
            sage: GF = I.groebner_fan()
            sage: PF = GF.tropical_intersection()
            sage: pfi0 = PF.initial_form_systems()[0]
            sage: pfi0.cone()
            [0]
        """
        return self._cone

    def rays(self):
        """
        The rays of the cone associated with the initial form system.

        EXAMPLES::

            sage: R.<x,y> = QQ[]
            sage: I = R.ideal([(x+y)^2-1,(x+y)^2-2,(x+y)^2-3])
            sage: GF = I.groebner_fan()
            sage: PF = GF.tropical_intersection()
            sage: pfi0 = PF.initial_form_systems()[0]
            sage: pfi0.rays()
            [[-1, 0]]
        """
        return self._rays

    def internal_ray(self):
        """
        A ray internal to the cone associated with the initial form
        system.

        EXAMPLES::

            sage: R.<x,y> = QQ[]
            sage: I = R.ideal([(x+y)^2-1,(x+y)^2-2,(x+y)^2-3])
            sage: GF = I.groebner_fan()
            sage: PF = GF.tropical_intersection()
            sage: pfi0 = PF.initial_form_systems()[0]
            sage: pfi0.internal_ray()
            (-1, 0)
        """
        return sum([vector(q) for q in self.rays()])

    def initial_forms(self):
        """
        The initial forms (polynomials).

        EXAMPLES::

            sage: R.<x,y> = QQ[]
            sage: I = R.ideal([(x+y)^2-1,(x+y)^2-2,(x+y)^2-3])
            sage: GF = I.groebner_fan()
            sage: PF = GF.tropical_intersection()
            sage: pfi0 = PF.initial_form_systems()[0]
            sage: pfi0.initial_forms()
            [y^2 - 1, y^2 - 2, y^2 - 3]
        """
        return self._initial_forms


def verts_for_normal(normal, poly) -> list:
    """
    Return the exponents of the vertices of a Newton polytope
    that make up the supporting hyperplane for the given outward
    normal.

    EXAMPLES::

        sage: from sage.rings.polynomial.groebner_fan import verts_for_normal
        sage: R.<x,y,z> = PolynomialRing(QQ,3)
        sage: f1 = x*y*z - 1
        sage: f2 = f1*(x^2 + y^2 + 1)
        sage: verts_for_normal([1,1,1],f2)
        [(3, 1, 1), (1, 3, 1)]
    """
    exps = [tuple(x) for x in poly.exponents()]
    expmat = matrix(exps)
    vals = expmat * vector(QQ, normal)
    maxval = max(vals)
    return [exps[i] for i in range(len(exps)) if vals[i] == maxval]


class TropicalPrevariety(PolyhedralFan):

    def __init__(self, gfan_polyhedral_fan, polynomial_system, poly_ring,
                 parameters=None) -> None:
        """
        This class is a subclass of the PolyhedralFan class,
        with some additional methods for tropical prevarieties.

        INPUT:

        - ``gfan_polyhedral_fan`` -- output from ``gfan`` of a polyhedral fan
        - ``polynomial_system`` -- list of polynomials
        - ``poly_ring`` -- the polynomial ring of the list of polynomials
        - ``parameters`` -- (optional) list of variables to be considered
          as parameters

        EXAMPLES::

            sage: R.<x,y,z> = QQ[]
            sage: I = R.ideal([(x+y+z)^2-1,(x+y+z)-x,(x+y+z)-3])
            sage: GF = I.groebner_fan()
            sage: TI = GF.tropical_intersection()
            sage: TI._polynomial_system
            [x^2 + 2*x*y + y^2 + 2*x*z + 2*y*z + z^2 - 1, y + z, x + y + z - 3]
        """
        parameter_indices = []
        if parameters is not None:
            allvars = poly_ring.gens()
            parameter_indices = [allvars.index(q) for q in parameters]
        PolyhedralFan.__init__(self, gfan_polyhedral_fan,
                               parameter_indices=parameter_indices)
        self._polynomial_system = polynomial_system
        self._parameters = parameters

    def initial_form_systems(self) -> list:
        """
        Return a list of systems of initial forms for each cone
        in the tropical prevariety.

        EXAMPLES::

            sage: R.<x,y> = QQ[]
            sage: I = R.ideal([(x+y)^2-1,(x+y)^2-2,(x+y)^2-3])
            sage: GF = I.groebner_fan()
            sage: PF = GF.tropical_intersection()
            sage: pfi = PF.initial_form_systems()
            sage: for q in pfi:
            ....:     print(q.initial_forms())
            [y^2 - 1, y^2 - 2, y^2 - 3]
            [x^2 - 1, x^2 - 2, x^2 - 3]
            [x^2 + 2*x*y + y^2, x^2 + 2*x*y + y^2, x^2 + 2*x*y + y^2]
        """
        try:
            return self._initial_form_systems
        except AttributeError:
            initial_form_systems = []
            pvars = self._polynomial_system[0].parent().gens()
            nvars = len(pvars)
            for dcone in self.cones():
                for acone in self.cones()[dcone]:
                    rays = [self.rays()[i] for i in acone]
                    repray = sum([vector(q) for q in rays])
                    iforms = []
                    for poly in self._polynomial_system:
                        verts = verts_for_normal(repray, poly)
                        nform = 0
                        for x in verts:
                            factorlist = [pvars[i]**x[i] for i in range(nvars)]
                            temp_monomial = prod(factorlist)
                            nform += poly.monomial_coefficient(temp_monomial) * temp_monomial
                        iforms.append(nform)
                    initial_form_systems.append(InitialForm(acone, rays, iforms))
            self._initial_form_systems = initial_form_systems
            return self._initial_form_systems


def ring_to_gfan_format(input_ring) -> str:
    """
    Convert a ring to gfan's format.

    EXAMPLES::

        sage: R.<w,x,y,z> = QQ[]
        sage: from sage.rings.polynomial.groebner_fan import ring_to_gfan_format
        sage: ring_to_gfan_format(R)
        'Q[w, x, y, z]'
        sage: R2.<x,y> = GF(2)[]
        sage: ring_to_gfan_format(R2)
        'Z/2Z[x, y]'
    """
    gens = str(input_ring.gens()).replace('(', '[').replace(')', ']')
    if input_ring.base_ring() is QQ:
        return 'Q' + gens
    if input_ring.base_ring() is ZZ:
        return 'Z' + gens
    return 'Z/{}Z'.format(input_ring.characteristic()) + gens


def ideal_to_gfan_format(input_ring, polys) -> str:
    """
    Return the ideal in gfan's notation.

        EXAMPLES::

            sage: R.<x,y,z> = PolynomialRing(QQ,3)
            sage: polys = [x^2*y - z, y^2*z - x, z^2*x - y]
            sage: from sage.rings.polynomial.groebner_fan import ideal_to_gfan_format
            sage: ideal_to_gfan_format(R, polys)
            'Q[x, y, z]{x^2*y-z,y^2*z-x,x*z^2-y}'

        TESTS:

        Test that :issue:`20146` is fixed::

            sage: P = PolynomialRing(QQ,"x11,x12,x13,x14,x15,x21,x22,x23,x24,x25,x31,x32,x33,x34,x35"); x = P.gens(); M = Matrix(3,x)
            sage: I = P.ideal(M.minors(2))
            sage: ideal_to_gfan_format(P,I.gens())
            'Q[x11, x12, x13, x14, x15, x21, x22, x23, x24, x25, x31, x32, x33, x34, x35]{-x12*x21+x11*x22,-x13*x21+x11*x23,-x14*x21+x11*x24,-x15*x21+x11*x25,-x13*x22+x12*x23,-x14*x22+x12*x24,-x15*x22+x12*x25,-x14*x23+x13*x24,-x15*x23+x13*x25,-x15*x24+x14*x25,-x12*x31+x11*x32,-x13*x31+x11*x33,-x14*x31+x11*x34,-x15*x31+x11*x35,-x13*x32+x12*x33,-x14*x32+x12*x34,-x15*x32+x12*x35,-x14*x33+x13*x34,-x15*x33+x13*x35,-x15*x34+x14*x35,-x22*x31+x21*x32,-x23*x31+x21*x33,-x24*x31+x21*x34,-x25*x31+x21*x35,-x23*x32+x22*x33,-x24*x32+x22*x34,-x25*x32+x22*x35,-x24*x33+x23*x34,-x25*x33+x23*x35,-x25*x34+x24*x35}'
    """
    ideal_gen_str = "{" + ",".join(str(poly).replace(" ", "").replace("'", "")
                                   for poly in polys) + "}"
    ring_str = ring_to_gfan_format(input_ring)
    output = ring_str + ideal_gen_str
    return output


class GroebnerFan(SageObject):

    def __init__(self, I, is_groebner_basis=False, symmetry=None, verbose=False) -> None:
        """
        This class is used to access capabilities of the program ``Gfan``.

        In addition to computing Groebner fans, ``Gfan`` can compute
        other things in tropical geometry such as tropical prevarieties.

        INPUT:

        - ``I`` -- ideal in a multivariate polynomial ring

        - ``is_groebner_basis`` -- boolean (default: ``False``); if
          ``True``, then I.gens() must be a Groebner basis with respect to the
          standard degree lexicographic term order

        - ``symmetry`` -- (default: ``None``) if not ``None``, describes
          symmetries of the ideal

        - ``verbose`` -- (default: ``False``) if ``True``, printout
          useful info during computations

        EXAMPLES::

            sage: R.<x,y,z> = QQ[]
            sage: I = R.ideal([x^2*y - z, y^2*z - x, z^2*x - y])
            sage: G = I.groebner_fan(); G
            Groebner fan of the ideal:
            Ideal (x^2*y - z, y^2*z - x, x*z^2 - y) of Multivariate Polynomial Ring in x, y, z over Rational Field

        Here is an example of the use of the tropical_intersection command, and then using the RationalPolyhedralFan
        class to compute the Stanley-Reisner ideal of the tropical prevariety::

            sage: R.<x,y,z> = QQ[]
            sage: I = R.ideal([(x+y+z)^3-1,(x+y+z)^3-x,(x+y+z)-3])
            sage: GF = I.groebner_fan()
            sage: PF = GF.tropical_intersection()
            sage: PF.rays()
            [[-1, 0, 0], [0, -1, 0], [0, 0, -1], [1, 1, 1]]
            sage: RPF = PF.to_RationalPolyhedralFan()
            sage: RPF.Stanley_Reisner_ideal(PolynomialRing(QQ,4,'A, B, C, D'))
            Ideal (A*B, A*C, B*C*D) of Multivariate Polynomial Ring in A, B, C, D over Rational Field
        """
        self.__is_groebner_basis = is_groebner_basis
        self.__symmetry = symmetry
        if symmetry:
            print("WARNING! Symmetry option not yet implemented!!")
        self.__verbose = verbose
        if not isinstance(I, MPolynomialIdeal):
            raise TypeError("I must be a multivariate polynomial ideal")
        if not prefix_check([str(R_gen) for R_gen in I.ring().gens()]):
            raise RuntimeError("Ring variables cannot contain each other as prefixes")
        S = I.ring()
        R = S.base_ring()
        # todo: add support for ZZ, which only works for bases computation, not tropical intersections
        if not R.is_field():
            raise NotImplementedError("Groebner fan computation only implemented over fields")
        if not (R is QQ or (R.is_finite() and R.is_prime_field() and R.order() <= 32749)):
            # 32749 is previous_prime(2^15)
            raise NotImplementedError("Groebner fan computation only implemented over Q or GF(p) for p <= 32749.")
        if S.ngens() > 52:
            raise NotImplementedError("Groebner fan computation only implemented for rings in at most 52 variables.")

        self.__ideal = I
        self.__ring = S

    def _repr_(self) -> str:
        """
        Describe the Groebner fan and its corresponding ideal.

        EXAMPLES::

            sage: R.<q,u> = PolynomialRing(QQ,2)
            sage: gf = R.ideal([q-u,u^2-1]).groebner_fan()
            sage: gf # indirect doctest
            Groebner fan of the ideal:
            Ideal (q - u, u^2 - 1) of Multivariate Polynomial Ring in q, u over Rational Field
        """
        return "Groebner fan of the ideal:\n{}".format(self.__ideal)

    def __eq__(self, right) -> bool:
        """
        Test equality of Groebner fan objects.

        EXAMPLES::

            sage: R.<q,u> = PolynomialRing(QQ,2)
            sage: gf = R.ideal([q^2-u,u^2-q]).groebner_fan()
            sage: gf2 = R.ideal([u^2-q,q^2-u]).groebner_fan()
            sage: gf.__eq__(gf2)
            True
        """
        return type(self) is type(right) and self.ideal() == right.ideal()

    def ideal(self):
        """
        Return the ideal the was used to define this Groebner fan.

        EXAMPLES::

            sage: R.<x1,x2> = PolynomialRing(QQ,2)
            sage: gf = R.ideal([x1^3-x2,x2^3-2*x1-2]).groebner_fan()
            sage: gf.ideal()
            Ideal (x1^3 - x2, x2^3 - 2*x1 - 2) of Multivariate Polynomial Ring in x1, x2 over Rational Field
        """
        return self.__ideal

    @cached_method
    def _gfan_maps(self) -> tuple:
        """
        OUTPUT:

        - map from Sage ring to ``gfan`` ring

        - map from ``gfan`` ring to Sage ring

        EXAMPLES::

            sage: R.<x,y,z> = PolynomialRing(QQ,3)
            sage: G = R.ideal([x^2*y - z, y^2*z - x, z^2*x - y]).groebner_fan()
            sage: G._gfan_maps()
            (Ring morphism:
              From: Multivariate Polynomial Ring in x, y, z over Rational Field
              To:   Multivariate Polynomial Ring in a, b, c over Rational Field
              Defn: x |--> a
                    y |--> b
                    z |--> c,
             Ring morphism:
              From: Multivariate Polynomial Ring in a, b, c over Rational Field
              To:   Multivariate Polynomial Ring in x, y, z over Rational Field
              Defn: a |--> x
                    b |--> y
                    c |--> z)
        """
        S = self.__ring
        n = S.ngens()

        # Define a polynomial ring in n variables
        # that are named a,b,c,d, ..., z, A, B, C, ...
        R = S.base_ring()
        T = PolynomialRing(R, n, string.ascii_letters[:n])

        # Define the homomorphism that sends the
        # generators of S to the generators of T.
        phi = S.hom(T.gens())

        # Define the homomorphism that sends the
        # generators of T to the generators of S.
        psi = T.hom(S.gens())
        return (phi, psi)

    def _gfan_ring(self) -> str:
        """
        Return the ring in ``gfan`` notation.

        EXAMPLES::

            sage: R.<x,y,z> = PolynomialRing(QQ,3)
            sage: G = R.ideal([x^2*y - z, y^2*z - x, z^2*x - y]).groebner_fan()
            sage: G._gfan_ring()
            'Q[x, y, z]'
        """
        return ring_to_gfan_format(self.ring())

    @cached_method
    def _gfan_ideal(self) -> str:
        """
        Return the ideal in ``gfan`` notation.

        EXAMPLES::

            sage: R.<x,y,z> = PolynomialRing(QQ,3)
            sage: G = R.ideal([x^2*y - z, y^2*z - x, z^2*x - y]).groebner_fan()
            sage: G._gfan_ideal()
            'Q[x, y, z]{x^2*y-z,y^2*z-x,x*z^2-y}'
        """
        return ideal_to_gfan_format(self.ring(),
                                    self.__ideal.gens())

    def weight_vectors(self) -> list:
        """
        Return the weight vectors corresponding to the reduced Groebner
        bases.

        EXAMPLES::

            sage: r3.<x,y,z> = PolynomialRing(QQ,3)
            sage: g = r3.ideal([x^3+y,y^3-z,z^2-x]).groebner_fan()
            sage: g.weight_vectors()
            [(3, 7, 1), (5, 1, 2), (7, 1, 4), (5, 1, 4), (1, 1, 1), (1, 4, 8), (1, 4, 10)]
            sage: r4.<x,y,z,w> = PolynomialRing(QQ,4)
            sage: g4 = r4.ideal([x^3+y,y^3-z,z^2-x,z^3 - w]).groebner_fan()
            sage: len(g4.weight_vectors())
            23
        """
        gfan_processes = Popen(['gfan', '_weightvector', '-m'],
                               stdin=PIPE, stdout=PIPE, stderr=PIPE)
        b_ans, _ = gfan_processes.communicate(input=self.gfan().encode("utf8"))
        s_ans = b_ans.decode()
        vect = re.compile(r"\([0-9,/\s]*\)")
        ans = (tup[1:-1].split(',') for tup in vect.findall(s_ans))
        return [vector(QQ, [QQ(y) for y in x]) for x in ans]

    def ring(self):
        """
        Return the multivariate polynomial ring.

        EXAMPLES::

            sage: R.<x1,x2> = PolynomialRing(QQ,2)
            sage: gf = R.ideal([x1^3-x2,x2^3-x1-2]).groebner_fan()
            sage: gf.ring()
            Multivariate Polynomial Ring in x1, x2 over Rational Field
        """
        return self.__ring

    @cached_method
    def _gfan_reduced_groebner_bases(self) -> str:
        """
        A string of the reduced Groebner bases of the ideal as output by
        ``gfan``.

        EXAMPLES::

            sage: R.<a,b> = PolynomialRing(QQ,2)
            sage: gf = R.ideal([a^3-b^2,b^2-a-1]).groebner_fan()
            sage: gf._gfan_reduced_groebner_bases()
            'Q[a,b]{{b^6-1+2*b^2-3*b^4,a+1-b^2},{a^3-1-a,b^2-1-a}}'
        """
        B = self.gfan(cmd='bases')
        B = B.replace('\n', '')
        return B

    def characteristic(self):
        """
        Return the characteristic of the base ring.

        EXAMPLES::

            sage: R.<x,y,z> = PolynomialRing(QQ,3)
            sage: i1 = ideal(x*z + 6*y*z - z^2, x*y + 6*x*z + y*z - z^2, y^2 + x*z + y*z)
            sage: gf = i1.groebner_fan()
            sage: gf.characteristic()
            0
        """
        return self.__ring.characteristic()

    @cached_method
    def reduced_groebner_bases(self):
        """
        EXAMPLES::

            sage: R.<x,y,z> = PolynomialRing(QQ, 3, order='lex')
            sage: G = R.ideal([x^2*y - z, y^2*z - x, z^2*x - y]).groebner_fan()
            sage: X = G.reduced_groebner_bases()
            sage: len(X)
            33
            sage: X[0]
            [z^15 - z, x - z^9, y - z^11]
            sage: X[0].ideal()
            Ideal (z^15 - z, x - z^9, y - z^11) of Multivariate Polynomial Ring in x, y, z over Rational Field
            sage: X[:5]
            [[z^15 - z, x - z^9, y - z^11],
            [y^2 - z^8, x - z^9, y*z^4 - z, -y + z^11],
            [y^3 - z^5, x - y^2*z, y^2*z^3 - y, y*z^4 - z, -y^2 + z^8],
            [y^4 - z^2, x - y^2*z, y^2*z^3 - y, y*z^4 - z, -y^3 + z^5],
            [y^9 - z, y^6*z - y, x - y^2*z, -y^4 + z^2]]
            sage: R3.<x,y,z> = PolynomialRing(GF(2477),3)
            sage: gf = R3.ideal([300*x^3-y,y^2-z,z^2-12]).groebner_fan()
            sage: gf.reduced_groebner_bases()
            [[z^2 - 12, y^2 - z, x^3 + 933*y],
            [y^4 - 12, x^3 + 933*y, -y^2 + z],
            [x^6 - 1062*z, z^2 - 12, -300*x^3 + y],
            [x^12 + 200, -300*x^3 + y, -828*x^6 + z]]
        """
        G = self._gfan_reduced_groebner_bases()
        if G.find(']') != -1:
            G = G.split(']')[1]
        G = G.replace('{{', '').replace('}}', '').split('},{')
        S = self.__ring
        return [ReducedGroebnerBasis(self, [S(f) for f in G[i].split(',')],
                                     G[i]) for i in range(len(G))]

    @cached_method
    def _gfan_mod(self) -> str:
        """
        Return the extra options to the ``gfan`` command that are used
        by this object to account for working modulo a prime or in the
        presence of extra symmetries.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(QQ,2)
            sage: gf = R.ideal([x^3-y,y^3-x-1]).groebner_fan()
            sage: gf._gfan_mod()
            ''
        """
        mod = ''
        # p = self.characteristic()
        # if p:
        #     mod += ' --mod %s' % p
        # else:
        #     mod += ''

        if self.__is_groebner_basis:
            mod += ' -g'

        if self.__symmetry:
            mod += ' --symmetry'

        return mod

    def gfan(self, cmd='bases', I=None) -> str:
        r"""
        Return the ``gfan`` output as a string given an input ``cmd``.

        The default is to produce the list of reduced Groebner bases
        in ``gfan`` format.

        INPUT:

        - ``cmd`` -- string (default: ``'bases'``); GFan command
        - ``I`` -- ideal (default: ``None``)

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(QQ,2)
            sage: gf = R.ideal([x^3-y,y^3-x-1]).groebner_fan()
            sage: gf.gfan()
            'Q[x,y]\n{{\ny^9-1-y+3*y^3-3*y^6,\nx+1-y^3}\n,\n{\nx^3-y,\ny^3-1-x}\n,\n{\nx^9-1-x,\ny-x^3}\n}\n'
        """
        if I is None:
            I = self._gfan_ideal()
        # todo -- put something in here (?) when self.__symmetry isn't None...
        cmd += self._gfan_mod()
        s = gfan(I, cmd, verbose=self.__verbose)
        if s.strip() == '{':
            raise RuntimeError("Error running gfan command %s on %s" % (cmd, self))
        return s

    def __iter__(self) -> Iterator:
        """
        Return an iterator for the reduced Groebner bases.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(QQ,2)
            sage: gf = R.ideal([x^3-y,y^3-x-1]).groebner_fan()
            sage: a = gf.__iter__()
            sage: next(a)
            [y^9 - 3*y^6 + 3*y^3 - y - 1, -y^3 + x + 1]
        """
        yield from self.reduced_groebner_bases()

    def __getitem__(self, i):
        """
        Get a reduced groebner basis.

        EXAMPLES::

            sage: R4.<w1,w2,w3,w4> = PolynomialRing(QQ,4)
            sage: gf = R4.ideal([w1^2-w2,w2^3-1,2*w3-w4^2,w4^2-w1]).groebner_fan()
            sage: gf[0]
            [w4^12 - 1, -w4^4 + w2, -w4^2 + w1, -1/2*w4^2 + w3]
        """
        return self.reduced_groebner_bases()[i]

    @cached_method
    def buchberger(self):
        """
        Return a lexicographic reduced Groebner basis for the ideal.

        EXAMPLES::

            sage: R.<x,y,z> = PolynomialRing(QQ,3)
            sage: G = R.ideal([x - z^3, y^2 - x + x^2 - z^3*x]).groebner_fan()
            sage: G.buchberger()
            [-z^3 + y^2, -z^3 + x]
        """
        B = self.gfan(cmd='buchberger')
        if B.find(']') != -1:
            B = B.split(']')[1]
        B = B.replace('}', '').replace('{', '')
        S = self.__ring
        return [S(f) for f in B.split(',')]

    @cached_method
    def polyhedralfan(self):
        """
        Return a polyhedral fan object corresponding to the reduced
        Groebner bases.

        EXAMPLES::

            sage: R3.<x,y,z> = PolynomialRing(QQ,3)
            sage: gf = R3.ideal([x^8-y^4,y^4-z^2,z^2-1]).groebner_fan()
            sage: pf = gf.polyhedralfan()
            sage: pf.rays()
            [[0, 0, 1], [0, 1, 0], [1, 0, 0]]
        """
        f = self.gfan(cmd='topolyhedralfan', I=self._gfan_reduced_groebner_bases())
        return PolyhedralFan(f)

    @cached_method
    def homogeneity_space(self):
        """
        Return the homogeneity space of a the list of polynomials that
        define this Groebner fan.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(QQ,2)
            sage: G = R.ideal([y^3 - x^2, y^2 - 13*x]).groebner_fan()
            sage: H = G.homogeneity_space()
        """
        return self.gfan(cmd='homogeneityspace')

    def render(self, file=None, larger=False, shift=0, rgbcolor=(0, 0, 0),
               polyfill=True, scale_colors=True):
        """
        Render a Groebner fan as sage graphics or save as an xfig file.

        More precisely, the output is a drawing of the Groebner fan
        intersected with a triangle. The corners of the triangle are
        (1,0,0) to the right, (0,1,0) to the left and (0,0,1) at the top.
        If there are more than three variables in the ring we extend these
        coordinates with zeros.

        INPUT:

        - ``file`` -- a filename if you prefer the output
          saved to a file; this will be in xfig format

        - ``shift`` -- shift the positions of the variables in
          the drawing. For example, with shift=1, the corners will be b
          (right), c (left), and d (top). The shifting is done modulo the
          number of variables in the polynomial ring. The default is 0.

        - ``larger`` -- boolean (default: ``False``); if ``True``, make
          the triangle larger so that the shape of the Groebner region
          appears. Affects the xfig file but probably not the sage graphics (?).

        - ``rgbcolor`` -- this will not affect the saved xfig
          file, only the sage graphics produced

        - ``polyfill`` -- whether or not to fill the cones with
          a color determined by the highest degree in each reduced Groebner
          basis for that cone

        - ``scale_colors`` -- if ``True``, this will normalize
          color values to try to maximize the range

        EXAMPLES::

            sage: R.<x,y,z> = PolynomialRing(QQ,3)
            sage: G = R.ideal([y^3 - x^2, y^2 - 13*x,z]).groebner_fan()
            sage: test_render = G.render()                                              # needs sage.plot

        ::

            sage: R.<x,y,z> = PolynomialRing(QQ,3)
            sage: G = R.ideal([x^2*y - z, y^2*z - x, z^2*x - y]).groebner_fan()
            sage: test_render = G.render(larger=True)                                   # needs sage.plot

        TESTS:

        Testing the case where the number of generators is < 3. Currently,
        this should raise a :exc:`NotImplementedError`.

        ::

            sage: R.<x,y> = PolynomialRing(QQ, 2)
            sage: R.ideal([y^3 - x^2, y^2 - 13*x]).groebner_fan().render()              # needs sage.plot
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        if polyfill is True:
            polyfill = max_degree
        S = self.__ring
        if S.ngens() < 3:
            print("For 2-D fan rendering the polynomial ring must have 3 variables (or more, which are ignored).")
            raise NotImplementedError
        cmd = 'render'
        if shift:
            cmd += ' --shiftVariables %s' % shift
        if larger:
            cmd += ' -L'
        s = self.gfan(cmd, I=self._gfan_reduced_groebner_bases().replace(' ', ','))
        if file is not None:
            with open(file, 'w') as f:
                f.write(s)
        sp = s.split('\n')
        sp2 = []
        for x in sp[9:]:
            xs = x.split(' ')
            y = []
            if x[0:3] != '2 3' and len(xs) > 1:
                y.extend(q for q in xs if q)
                sp2.append(y)
        sp3 = []
        for j in range(len(sp2)):
            temp = [[float(sp2[j][i]) / 1200.0,
                     float(sp2[j][i + 1]) / 1200.0]
                    for i in range(0, len(sp2[j]) - 1, 2)]
            sp3.append(temp)
        r_lines = Graphics()
        for x in sp3:
            r_lines = r_lines + line(x, rgbcolor=rgbcolor)
        if polyfill:
            vals = [polyfill(q) for q in self.reduced_groebner_bases()]
            if isinstance(vals[0], (list, tuple)):
                if scale_colors:
                    vmins = [min([q[i] for q in vals]) for i in (0, 1, 2)]
                    vmaxs = [max([q[i] for q in vals]) for i in (0, 1, 2)]
                    for i in (0, 1, 2):
                        if vmaxs[i] == vmins[i]:
                            vmaxs[i] = vmins[i] + .01
                    for index, sp in enumerate(sp3):
                        col = [1 - (vals[index][i] - vmins[i]) / (vmaxs[i] - vmins[i]) for i in (0, 1, 2)]
                        r_lines += polygon(sp, rgbcolor=col)
                else:
                    for index, sp in enumerate(sp3):
                        r_lines += polygon(sp, rgbcolor=vals[index])
            elif scale_colors:
                vmin = min(vals)
                vmax = max(vals)
                if vmin == vmax:
                    vmax = vmin + .01
                for index, sp in enumerate(sp3):
                    r_lines += polygon(sp, hue=.1 + .6 * (vals[index] - vmin) / (vmax - vmin))
            else:
                for index, sp in enumerate(sp3):
                    r_lines += polygon(sp, hue=vals[index])
        return r_lines

    def _cone_to_ieq(self, facet_list):
        """
        A simple utility function for converting a facet normal to an
        inequality form.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(QQ,2) # dummy stuff to get a gfan object
            sage: gf = R.ideal([x^2+y,y^2]).groebner_fan()
            sage: gf._cone_to_ieq([[1,2,3,4]])
            [[0, 1, 2, 3, 4]]
        """
        return [[0] + q for q in facet_list]

    def _embed_tetra(self, fpoint):
        """
        Take a 4-d vector and projects it onto the plane perpendicular to
        (1,1,1,1). Stretches by a factor of 2 as well, since this is only
        for graphical display purposes.

        INPUT:

        - ``fpoint`` -- list of four numbers

        EXAMPLES::

            sage: R.<x> = PolynomialRing(QQ,1) # dummy stuff to get a gfan object
            sage: gf = R.ideal([x^2]).groebner_fan()
            sage: gf._embed_tetra([1/2,1/2,1/2,1/2])
            [0, 0, 0]
        """
        v1 = [1, 1, -1, -1]
        v2 = [1, -1, 1, -1]
        v3 = [-1, 1, 1, -1]
        x1 = sum([fpoint[ind] * v1[ind] for ind in range(4)])
        x2 = sum([fpoint[ind] * v2[ind] for ind in range(4)])
        x3 = sum([fpoint[ind] * v3[ind] for ind in range(4)])
        return [x1, x2, x3]

    def _4d_to_3d(self, polyhedral_data):
        """
        A utility function that takes a list of 4d polytopes, projects them
        to 3d, and returns a list of edges.

        INPUT:

        - ``polyhedral_data`` -- an object with 4d vertex and adjacency
          information

        OUTPUT:

        - ``edges`` -- list of edges in 3d; each list item is a pair of
          points

        EXAMPLES::

            sage: R4.<w,x,y,z> = PolynomialRing(QQ,4)
            sage: gf = R4.ideal([w^2-x,x^2-y,y^2-z,z^2-1]).groebner_fan()
            sage: g_cone = gf[0].groebner_cone()
            sage: g_cone_facets = g_cone.facets()
            sage: g_cone_ieqs = gf._cone_to_ieq(g_cone_facets)
            sage: cone_data = Polyhedron(ieqs = g_cone_ieqs, eqns = [[1,-1,-1,-1,-1]])
            sage: cone_lines = gf._4d_to_3d(cone_data)
            sage: cone_lines
            [[[1, 1, -1], [1, -1/3, 1/3]],
            [[1, 1, -1], [-1/7, 3/7, 5/7]],
            [[1, 1, -1], [-3/5, -1/3, -1/5]],
            [[1, -1/3, 1/3], [-1/7, 3/7, 5/7]],
            [[1, -1/3, 1/3], [-3/5, -1/3, -1/5]],
            [[-1/7, 3/7, 5/7], [-3/5, -1/3, -1/5]]]
        """
        fpoints = polyhedral_data.vertices()
        tpoints = [self._embed_tetra(q) for q in fpoints]
        edges = []
        for vertex in polyhedral_data.vertices():
            i = vertex.index()
            for adjacent_vertex in vertex.adjacent():
                j = adjacent_vertex.index()
                if j > i:
                    try:
                        edges.append([tpoints[i], tpoints[j]])
                    except Exception:
                        print('tpoints: ' + str(tpoints))
                        print('fpoints: ' + str(fpoints))
                        print(adjacent_vertex)
                        print(polyhedral_data.ieqs())
                        raise RuntimeError
        return edges

    def render3d(self, verbose=False):
        """
        For a Groebner fan of an ideal in a ring with four variables, this
        function intersects the fan with the standard simplex perpendicular
        to (1,1,1,1), creating a 3d polytope, which is then projected into
        3 dimensions. The edges of this projected polytope are returned as
        lines.

        EXAMPLES::

            sage: R4.<w,x,y,z> = PolynomialRing(QQ,4)
            sage: gf = R4.ideal([w^2-x,x^2-y,y^2-z,z^2-x]).groebner_fan()
            sage: three_d = gf.render3d()                                               # needs sage.plot

        TESTS:

        Now test the case where the number of generators is not 4. Currently,
        this should raise a :exc:`NotImplementedError` error.

        ::

            sage: P.<a,b,c> = PolynomialRing(QQ, 3, order='lex')
            sage: sage.rings.ideal.Katsura(P, 3).groebner_fan().render3d()              # needs sage.plot
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        S = self.__ring
        if S.ngens() != 4:
            print("For 3-D fan rendering the polynomial ring must have 4 variables")
            raise NotImplementedError
        g_cones = [q.groebner_cone() for q in self.reduced_groebner_bases()]
        g_cones_facets = [q.facets() for q in g_cones]
        g_cones_ieqs = [self._cone_to_ieq(q) for q in g_cones_facets]
        # Now the cones are intersected with a plane:
        cone_info = [Polyhedron(ieqs=q, eqns=[[1, -1, -1, -1, -1]])
                     for q in g_cones_ieqs]
        # This is really just for debugging
        if verbose:
            for x in cone_info:
                print(x.inequalities() + ([1, 1, 0, 0, 0], [1, 0, 1, 0, 0],
                                          [1, 0, 0, 1, 0], [1, 0, 0, 0, 1]))
                print(x.equations())
                print()
        cone_info = [Polyhedron(ieqs=x.inequalities() +
                                ([1, 1, 0, 0, 0], [1, 0, 1, 0, 0],
                                 [1, 0, 0, 1, 0], [1, 0, 0, 0, 1]),
                                eqns=x.equations()) for x in cone_info]
        all_lines = []
        for cone_data in cone_info:
            try:
                cone_lines = self._4d_to_3d(cone_data)
            except Exception:
                print(cone_data._rays)
                raise RuntimeError
            all_lines.extend(a_line for a_line in cone_lines)
        return sum([line3d(a_line) for a_line in all_lines])

    @cached_method
    def _gfan_stats(self) -> dict:
        """
        Return various statistics about this Groebner fan.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(QQ)
            sage: G = R.ideal([y^3 - x^2, y^2 - 13*x]).groebner_fan()
            sage: G._gfan_stats()
            {'Dimension of homogeneity space': 0,
             'Maximal number of polynomials in Groebner basis': 3,
             'Maximal number of terms in Groebner basis': 6,
             'Maximal total degree of a Groebner basis': 4,
             'Minimal total degree of a Groebner basis': 2,
             'Number of reduced Groebner bases': 3,
             'Number of variables': 2}
        """
        s = self.gfan(cmd='stats',
                      I=self._gfan_reduced_groebner_bases().replace(' ', ','))
        d = {}
        for v in s.split('\n'):
            if v:
                a, b = v.split(':')
                d[a] = ZZ(b)
        return d

    def dimension_of_homogeneity_space(self):
        """
        Return the dimension of the homogeneity space.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(QQ,2)
            sage: G = R.ideal([y^3 - x^2, y^2 - 13*x]).groebner_fan()
            sage: G.dimension_of_homogeneity_space()
            0
        """
        return self._gfan_stats()['Dimension of homogeneity space']

    def maximal_total_degree_of_a_groebner_basis(self):
        """
        Return the maximal total degree of any Groebner basis.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(QQ,2)
            sage: G = R.ideal([y^3 - x^2, y^2 - 13*x]).groebner_fan()
            sage: G.maximal_total_degree_of_a_groebner_basis()
            4
        """
        return self._gfan_stats()['Maximal total degree of a Groebner basis']

    def minimal_total_degree_of_a_groebner_basis(self):
        """
        Return the minimal total degree of any Groebner basis.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(QQ,2)
            sage: G = R.ideal([y^3 - x^2, y^2 - 13*x]).groebner_fan()
            sage: G.minimal_total_degree_of_a_groebner_basis()
            2
        """
        return self._gfan_stats()['Minimal total degree of a Groebner basis']

    def number_of_reduced_groebner_bases(self):
        """
        Return the number of reduced Groebner bases.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(QQ,2)
            sage: G = R.ideal([y^3 - x^2, y^2 - 13*x]).groebner_fan()
            sage: G.number_of_reduced_groebner_bases()
            3
        """
        return self._gfan_stats()['Number of reduced Groebner bases']

    def number_of_variables(self):
        """
        Return the number of variables.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(QQ,2)
            sage: G = R.ideal([y^3 - x^2, y^2 - 13*x]).groebner_fan()
            sage: G.number_of_variables()
            2

        ::

            sage: R = PolynomialRing(QQ,'x',10)
            sage: R.inject_variables(globals())
            Defining x0, x1, x2, x3, x4, x5, x6, x7, x8, x9
            sage: G = ideal([x0 - x9, sum(R.gens())]).groebner_fan()
            sage: G.number_of_variables()
            10
        """
        return self.__ring.ngens()

    @cached_method
    def tropical_basis(self, check=True, verbose=False):
        """
        Return a tropical basis for the tropical curve associated to this
        ideal.

        INPUT:

        - ``check`` -- boolean (default: ``True``); if ``True`` raises a
          :exc:`ValueError` exception if this ideal does not define a tropical
          curve (i.e., the condition that R/I has dimension equal to 1 + the
          dimension of the homogeneity space is not satisfied)

        EXAMPLES::

            sage: R.<x,y,z> = PolynomialRing(QQ,3, order='lex')
            sage: G = R.ideal([y^3-3*x^2, z^3-x-y-2*y^3+2*x^2]).groebner_fan()
            sage: G
            Groebner fan of the ideal:
            Ideal (-3*x^2 + y^3, 2*x^2 - x - 2*y^3 - y + z^3) of Multivariate Polynomial Ring in x, y, z over Rational Field
            sage: G.tropical_basis()
            [-3*x^2 + y^3, 2*x^2 - x - 2*y^3 - y + z^3, 3/4*x + y^3 + 3/4*y - 3/4*z^3]
        """
        cmd = 'tropicalbasis'

        I = self.ideal()
        if not I.is_homogeneous():
            cmd += ' -h'
        if check:
            if I.dimension() != 1 + self.dimension_of_homogeneity_space():
                raise ValueError("The ideal does not define a tropical curve.")

        B = self.gfan(cmd)
        if B.find(']') != -1:
            B = B.split(']')[1]
        S = self.__ring
        B = B.replace('\n', '')
        B = B.replace('{', '').replace('}', '').split(',')
        if verbose:
            print(S, B)
        return [S(f) for f in B]

    def interactive(self, *args, **kwds):
        """
        See the documentation for self[0].interactive().

        This does not work with the notebook.

        EXAMPLES::

            sage: print("This is not easily doc-testable; please write a good one!")
            This is not easily doc-testable; please write a good one!
        """
        self[0].interactive(*args, **kwds)

    @cached_method
    def tropical_intersection(self, parameters=None, symmetry_generators=None,
                              *args, **kwds):
        """
        Return information about the tropical intersection of the
        polynomials defining the ideal.

        This is the common refinement of the outward-pointing normal
        fans of the Newton polytopes of the generators of the
        ideal. Note that some people use the inward-pointing normal
        fans.

        INPUT:

        - ``parameters`` -- (optional) tuple of variables to be
          considered as parameters
        - ``symmetry_generators`` -- (optional) generators of the symmetry group

        OUTPUT: a TropicalPrevariety object

        EXAMPLES::

            sage: R.<x,y,z> = PolynomialRing(QQ,3)
            sage: I = R.ideal(x*z + 6*y*z - z^2, x*y + 6*x*z + y*z - z^2, y^2 + x*z + y*z)
            sage: gf = I.groebner_fan()
            sage: pf = gf.tropical_intersection()
            sage: pf.rays()
            [[-2, 1, 1]]

            sage: R.<x,y,z> = PolynomialRing(QQ,3)
            sage: f1 = x*y*z - 1
            sage: f2 = f1*(x^2 + y^2 + z^2)
            sage: f3 = f2*(x + y + z - 1)
            sage: I = R.ideal([f1,f2,f3])
            sage: gf = I.groebner_fan()
            sage: pf = gf.tropical_intersection(symmetry_generators = '(1,2,0),(1,0,2)')
            sage: pf.rays()
            [[-2, 1, 1], [1, -2, 1], [1, 1, -2]]

            sage: R.<x,y,z> = QQ[]
            sage: I = R.ideal([(x+y+z)^2-1,(x+y+z)-x,(x+y+z)-3])
            sage: GF = I.groebner_fan()
            sage: TI = GF.tropical_intersection()
            sage: TI.rays()
            [[-1, 0, 0], [0, -1, -1], [1, 1, 1]]
            sage: GF = I.groebner_fan()
            sage: TI = GF.tropical_intersection(parameters=(y,))
            sage: TI.rays()
            [[-1, 0, 0]]
        """
        if parameters is None:
            parameters = []
        if symmetry_generators is None:
            symmetry_generators = []
        cmd = 'tropicalintersection'
        id_str = self._gfan_ideal()
        if parameters:
            allvars = self.ring().gens()
            truevars = [q for q in allvars if q not in parameters]
            base_ring = self.ring().base_ring()
            new_ring = PolynomialRing(base_ring, len(truevars),
                                      ",".join(str(q) for q in truevars))
            old_polys = self.ideal().gens()
            new_polys = []
            sub = {v: 1 for v in parameters}
            for apoly in old_polys:
                mons = apoly.monomials()
                mons = [m.subs(sub) for m in mons]
                new_polys.append(sum(mons))
            id_str = ideal_to_gfan_format(new_ring, new_polys)
        if symmetry_generators:
            cmd = cmd + ' --symmetryExploit'
            id_str = id_str + '{' + symmetry_generators + '}'
        f = self.gfan(cmd=cmd, I=id_str)
        pf = TropicalPrevariety(f, self.ideal().gens(), self.ring(),
                                parameters=parameters)
        pf._gfan_output = f
        return pf

    def mixed_volume(self) -> Integer:
        """
        Return the mixed volume of the generators of this ideal.

        This is not really an ideal property, it can depend on the
        generators used.

        The generators must give a square system (as many polynomials
        as variables).

        EXAMPLES::

            sage: R.<x,y,z> = QQ[]
            sage: example_ideal = R.ideal([x^2-y-1,y^2-z-1,z^2-x-1])
            sage: gf = example_ideal.groebner_fan()
            sage: mv = gf.mixed_volume()
            sage: mv
            8

            sage: R2.<x,y> = QQ[]
            sage: g1 = 1 - x + x^7*y^3 + 2*x^8*y^4
            sage: g2 = 2 + y + 3*x^7*y^3 + x^8*y^4
            sage: example2 = R2.ideal([g1,g2])
            sage: example2.groebner_fan().mixed_volume()
            15
        """
        if len(self.ring().variable_names()) != self.ideal().ngens():
            raise ValueError('not a square system')
        return Integer(self.gfan(cmd='mixedvolume'))


class ReducedGroebnerBasis(SageObject, list):
    def __init__(self, groebner_fan, gens, gfan_gens) -> None:
        """
        A class for representing reduced Groebner bases as produced by
        ``gfan``.

        INPUT:

        - ``groebner_fan`` -- a GroebnerFan object from an ideal

        - ``gens`` -- the generators of the ideal

        - ``gfan_gens`` -- the generators as a gfan string

        EXAMPLES::

            sage: R.<a,b> = PolynomialRing(QQ,2)
            sage: gf = R.ideal([a^2-b^2,b-a-1]).groebner_fan()
            sage: from sage.rings.polynomial.groebner_fan import ReducedGroebnerBasis
            sage: ReducedGroebnerBasis(gf,gf[0],gf[0]._gfan_gens())
            [b - 1/2, a + 1/2]
        """
        self.__groebner_fan = groebner_fan
        list.__init__(self, gens)
        self.__gfan_gens = '{' + gfan_gens.replace(' ', ',') + '}'
        self.__ring = groebner_fan._gfan_ring()

    def _repr_(self) -> str:
        """
        Return the reduced Groebner basis as a string.

        EXAMPLES::

            sage: R.<z1,zz1> = PolynomialRing(QQ,2)
            sage: gf = R.ideal([z1^2*zz1-1,zz1-2]).groebner_fan()
            sage: rgb1 = gf.reduced_groebner_bases()[0]
            sage: rgb1 # indirect doctest
            [zz1 - 2, z1^2 - 1/2]
        """
        return list.__repr__(self)

    def _gfan_gens(self) -> str:
        """
        Return the reduced Groebner basis as a string in ``gfan`` format.

        EXAMPLES::

            sage: R.<z1,zz1> = PolynomialRing(QQ,2)
            sage: gf = R.ideal([z1^2*zz1-1,zz1-2]).groebner_fan()
            sage: rgb1 = gf.reduced_groebner_bases()[0]
            sage: rgb1._gfan_gens()
            '{zz1-2,z1^2-1/2}'
        """
        return self.__gfan_gens

    def _gfan(self):
        """
        Return a description of the Groebner fan this basis was derived
        from.

        EXAMPLES::

            sage: R.<z1,zz1> = PolynomialRing(QQ,2)
            sage: gf = R.ideal([z1^2*zz1-1,zz1-2]).groebner_fan()
            sage: rgb1 = gf.reduced_groebner_bases()[0]
            sage: rgb1._gfan()
            Groebner fan of the ideal:
            Ideal (z1^2*zz1 - 1, zz1 - 2) of Multivariate Polynomial Ring in z1, zz1 over Rational Field
        """
        return self.__groebner_fan

    def interactive(self, latex=False, flippable=False, wall=False,
                    inequalities=False, weight=False):
        """
        Do an interactive walk of the Groebner fan starting at this reduced
        Groebner basis.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(QQ,2)
            sage: G = R.ideal([y^3 - x^2, y^2 - 13*x]).groebner_fan()
            sage: G[0].interactive()      # not tested
            Initializing gfan interactive mode
            *********************************************
            *     Press control-C to return to Sage     *
            *********************************************
            ....
        """
        cmd = 'gfan_interactive'
        if latex:
            cmd += ' -L'
        if flippable:
            cmd += ' -f'
        if wall:
            cmd += ' -w'
        if inequalities:
            cmd += ' -i'
        if weight:
            cmd += ' -W'
        cmd += self.__groebner_fan._gfan_mod()
        E = pexpect.spawn(cmd)
        print("Initializing gfan interactive mode")
        # E.sendline(self._gfan_ideal())
        E.sendline(self.__gfan_gens)
        print("*" * 45)
        print("*     Press control-C to return to Sage     *")
        print("*" * 45)
        try:
            E.interact()
        except OSError:
            print("Returning to Sage.")

    def groebner_cone(self, restrict=False):
        """
        Return defining inequalities for the full-dimensional Groebner cone
        associated to this marked minimal reduced Groebner basis.

        INPUT:

        - ``restrict`` -- boolean (default: ``False``); if ``True``, add
          an inequality for each coordinate, so that the cone is restricted
          to the positive orthant

        OUTPUT: tuple of integer vectors

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(QQ,2)
            sage: G = R.ideal([y^3 - x^2, y^2 - 13*x]).groebner_fan()
            sage: poly_cone = G[1].groebner_cone()
            sage: poly_cone.facets()
            [[-1, 2], [1, -1]]
            sage: [g.groebner_cone().facets() for g in G]
            [[[0, 1], [1, -2]], [[-1, 2], [1, -1]], [[-1, 1], [1, 0]]]
            sage: G[1].groebner_cone(restrict=True).facets()
            [[-1, 2], [1, -1]]
        """
        try:
            return self.__groebner_cone[restrict]
        except AttributeError:
            self.__groebner_cone = {}
        except KeyError:
            pass
        cmd = 'groebnercone'
        if restrict:
            cmd += ' --restrict'
        gf = self.__groebner_fan
        c = gf.gfan(cmd=cmd, I=self.__ring + self.__gfan_gens)
        return PolyhedralCone(c)

    def ideal(self):
        """
        Return the ideal generated by this basis.

        EXAMPLES::

            sage: R.<x,y,z> = PolynomialRing(QQ,3)
            sage: G = R.ideal([x - z^3, y^2 - 13*x]).groebner_fan()
            sage: G[0].ideal()
            Ideal (-13*z^3 + y^2, -z^3 + x) of Multivariate Polynomial Ring in x, y, z over Rational Field
        """
        return self.__groebner_fan.ring().ideal(self)



# sage.doctest: needs sage.rings.number_field
r"""
Class functions of groups.

This module implements a wrapper of GAP's ClassFunction function.

This now goes through the libGAP library.

.. NOTE::

    The ordering of the columns of the character table of a group
    corresponds to the ordering of the list. However, in general there is
    no way to canonically list (or index) the conjugacy classes of a group.
    Therefore the ordering of  the columns of the character table of
    a group is somewhat random.

EXAMPLES::

    sage: G = CyclicPermutationGroup(4)
    sage: G.conjugacy_classes()
    [Conjugacy class of () in Cyclic group of order 4 as a permutation group,
     Conjugacy class of (1,2,3,4) in Cyclic group of order 4 as a permutation group,
     Conjugacy class of (1,3)(2,4) in Cyclic group of order 4 as a permutation group,
     Conjugacy class of (1,4,3,2) in Cyclic group of order 4 as a permutation group]
    sage: values  = [1, -1, 1, -1]
    sage: chi = ClassFunction(G, values); chi
    Character of Cyclic group of order 4 as a permutation group

TESTS::

    sage: from sage.groups.class_function import ClassFunction_libgap
    sage: ClassFunction_libgap(SymmetricGroup(2),[1,1])
    doctest:warning...ClassFunction_libgap is deprecated. Please use sage.groups.class_function.ClassFunction instead.
    See https://github.com/sagemath/sage/issues/41579 for details.
    Character of Symmetric group of order 2! as a permutation group

AUTHORS:

- Franco Saliola (November 2008): initial version

- Volker Braun (October 2010): Bugfixes, exterior and symmetric power.
"""

# ****************************************************************************
#       Copyright (C) 2008 Franco Saliola <saliola@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from __future__ import annotations

from sage.structure.sage_object import SageObject
from sage.structure.richcmp import richcmp, richcmp_method
from sage.rings.integer import Integer
from sage.libs.gap.libgap import libgap
from sage.libs.gap.element import GapElement
from sage.misc.superseded import deprecated_function_alias
from sage.misc.lazy_import import lazy_import

lazy_import('sage.rings.number_field.number_field', 'CyclotomicField')

# TODO:
#
# This module needs to be rewritten to implement the ring of class
# functions in the usual parent/element pattern. But
# https://github.com/sagemath/sage/issues/14014 is already too long...


#####################################################################
#
#   Class function using the GAP library
#
#####################################################################


@richcmp_method
class ClassFunction(SageObject):
    """
    A wrapper of GAP's ``ClassFunction`` function.

    .. NOTE::

        It is *not* checked whether the given values describes a character,
        since GAP does not do this.

    EXAMPLES::

        sage: G = SO(3,3)
        sage: values  = [1, -1, -1, 1, 2]
        sage: chi = ClassFunction(G, values); chi
        Character of Special Orthogonal Group of degree 3 over Finite Field of size 3
        sage: loads(dumps(chi)) == chi
        True
    """

    def __init__(self, G, values) -> None:
        r"""
        Return the character of the group ``G`` with values given by the list
        values. The order of the values must correspond to the output of
        ``G.conjugacy_classes_representatives()``.

        EXAMPLES::

            sage: G = CyclicPermutationGroup(4)
            sage: values  = [1, -1, 1, -1]
            sage: chi = ClassFunction(G, values); chi
            Character of Cyclic group of order 4 as a permutation group
        """
        self._group = G
        if isinstance(values, GapElement) and values.IsClassFunction():
            self._gap_classfunction = values
        else:
            self._gap_classfunction = libgap.ClassFunction(G, list(values))
        e = self._gap_classfunction.Conductor().sage()
        self._base_ring = CyclotomicField(e)

    def _libgap_(self):
        r"""
        Return the underlying LibGAP element.

        EXAMPLES::

            sage: G = CyclicPermutationGroup(4)
            sage: values  = [1, -1, 1, -1]
            sage: chi = ClassFunction(G, values);  chi
            Character of Cyclic group of order 4 as a permutation group
            sage: type(chi)
            <class 'sage.groups.class_function.ClassFunction'>
            sage: libgap(chi)
            ClassFunction( CharacterTable( Group([ (1,2,3,4) ]) ), [ 1, -1, 1, -1 ] )
            sage: type(_)
            <class 'sage.libs.gap.element.GapElement_List'>
        """
        return self._gap_classfunction

    gap = _libgap_

    def _repr_(self) -> str:
        r"""
        Return a string representation.

        EXAMPLES::

            sage: G = SymmetricGroup(4)
            sage: values  = [1, -1, 1, 1, -1]
            sage: ClassFunction(G, values)
            Character of Symmetric group of order 4! as a permutation group
        """
        return "Character of %s" % repr(self._group)

    def __hash__(self) -> int:
        r"""
        TESTS::

            sage: G = SymmetricGroup(5)
            sage: chi1 = ClassFunction(G,[1,1,1,1,1,1,1])
            sage: d = {chi1:'trivial'}
        """
        return hash((self._group, tuple(self)))

    def __iter__(self):
        r"""
        Iterate through the values.

        A class function assigns values to each conjugacy class. This
        method iterates over the values, in the same order as the
        conjugacy classes of the group.

        EXAMPLES::

            sage: xi = ClassFunction(SymmetricGroup(4), [1, -1, 1, 1, -1])
            sage: list(xi)
            [1, -1, 1, 1, -1]
        """
        for v in self._gap_classfunction.List():
            yield v.sage(ring=self._base_ring)

    def __richcmp__(self, other, op) -> bool:
        r"""
        Rich comparison for class functions.

        Compares groups and then the values of the class function on the
        conjugacy classes.

        EXAMPLES::

            sage: G = PermutationGroup([[(1,2,3),(4,5)],[(3,4)]])
            sage: chi = G.character([1, 1, 1, 1, 1, 1, 1])
            sage: H = PermutationGroup([[(1,2,3),(4,5)]])
            sage: xi = H.character([1, 1, 1, 1, 1, 1])
            sage: chi == chi
            True
            sage: xi == xi
            True
            sage: xi == chi
            False
            sage: chi < xi
            False
            sage: xi < chi
            True
        """
        if isinstance(other, ClassFunction):
            return richcmp((self._group, self.values()),
                           (other._group, other.values()), op)
        return NotImplemented

    def __reduce__(self) -> tuple:
        r"""
        Add pickle support.

        EXAMPLES::

            sage: G = GL(2,7)
            sage: values = G.gap().CharacterTable().Irr()[2].List().sage()
            sage: chi = ClassFunction(G, values)
            sage: type(chi)
            <class 'sage.groups.class_function.ClassFunction'>
            sage: loads(dumps(chi)) == chi
            True
        """
        return ClassFunction, (self._group, self.values())

    def domain(self):
        r"""
        Return the domain of ``self``.

        OUTPUT: the underlying group of the class function

        EXAMPLES::

            sage: ClassFunction(SymmetricGroup(4), [1,-1,1,1,-1]).domain()
            Symmetric group of order 4! as a permutation group
        """
        return self._group

    def __call__(self, g):
        """
        Evaluate the character on the group element `g`.

        Return an error if `g` is not in `G`.

        EXAMPLES::

            sage: G = GL(2,7)
            sage: values = G.gap().CharacterTable().Irr()[2].List().sage()
            sage: chi = ClassFunction(G, values)
            sage: z = G([[3,0],[0,3]]); z
            [3 0]
            [0 3]
            sage: chi(z)
            zeta3

            sage: G = GL(2,3)
            sage: chi = G.irreducible_characters()[3]
            sage: g = G.conjugacy_classes_representatives()[6]
            sage: chi(g)
            zeta8^3 + zeta8

            sage: G = SymmetricGroup(3)
            sage: h = G((2,3))
            sage: triv = G.trivial_character()
            sage: triv(h)
            1
        """
        value = g.gap() ** self.gap()
        return value.sage(self._base_ring)

    def __add__(self, other):
        r"""
        Return the sum of the characters ``self`` and ``other``.

        INPUT:

        - ``other`` -- a :class:`ClassFunction` of the same group as ``self``

        OUTPUT: a :class:`ClassFunction`

        EXAMPLES::

            sage: chi = ClassFunction(SymmetricGroup(4), [3, 1, -1, 0, -1])
            sage: s = chi+chi
            sage: s
            Character of Symmetric group of order 4! as a permutation group
            sage: s.values()
            [6, 2, -2, 0, -2]
        """
        if not isinstance(other, ClassFunction):
            raise NotImplementedError
        s = self._gap_classfunction + other._gap_classfunction
        return ClassFunction(self._group, s)

    def __sub__(self, other):
        r"""
        Return the difference of the characters ``self`` and ``other``.

        INPUT:

        - ``other`` -- a :class:`ClassFunction` of the same group as ``self``

        OUTPUT: a :class:`ClassFunction`

        EXAMPLES::

            sage: G = SymmetricGroup(4)
            sage: chi1 = ClassFunction(G, [3, 1, -1, 0, -1])
            sage: chi2 = ClassFunction(G, [1, -1, 1, 1, -1])
            sage: s = chi1 - chi2
            sage: s
            Character of Symmetric group of order 4! as a permutation group
            sage: s.values()
            [2, 2, -2, -1, 0]
        """
        if not isinstance(other, ClassFunction):
            raise NotImplementedError
        s = self._gap_classfunction - other._gap_classfunction
        return ClassFunction(self._group, s)

    def __mul__(self, other):
        r"""
        Return the product of the character with ``other``.

        INPUT:

        - ``other`` -- either a number or a :class:`ClassFunction` of
          the same group as ``self``. A number can be anything that
          can be converted into GAP: integers, rational, and elements
          of certain number fields.

        OUTPUT: a :class:`ClassFunction`

        EXAMPLES::

            sage: G = SymmetricGroup(4)
            sage: chi1 = ClassFunction(G, [3, 1, -1, 0, -1])
            sage: 3*chi1
            Character of Symmetric group of order 4! as a permutation group
            sage: 3*chi1 == chi1+chi1+chi1
            True
            sage: (3*chi1).values()
            [9, 3, -3, 0, -3]

            sage: (1/2*chi1).values()
            [3/2, 1/2, -1/2, 0, -1/2]

            sage: CF3 = CyclotomicField(3)
            sage: CF3.inject_variables()
            Defining zeta3
            sage: (zeta3 * chi1).values()
            [3*zeta3, zeta3, -zeta3, 0, -zeta3]

            sage: chi2 = ClassFunction(G, [1, -1, 1, 1, -1])
            sage: p = chi1*chi2
            sage: p
            Character of Symmetric group of order 4! as a permutation group
            sage: p.values()
            [3, -1, -1, 0, 1]
        """
        if isinstance(other, ClassFunction):
            p = self._gap_classfunction * other._gap_classfunction
            return ClassFunction(self._group, p)
        return ClassFunction(self._group, other * self._gap_classfunction)

    def __rmul__(self, other):
        r"""
        Return the reverse multiplication of ``self`` and ``other``.

        EXAMPLES::

            sage: G = SymmetricGroup(4)
            sage: chi = ClassFunction(G, [3, 1, -1, 0, -1])
            sage: chi * 4   # calls chi.__mul__
            Character of Symmetric group of order 4! as a permutation group
            sage: 4 * chi   # calls chi.__rmul__
            Character of Symmetric group of order 4! as a permutation group
            sage: (4 * chi).values()
            [12, 4, -4, 0, -4]
        """
        return self.__mul__(other)

    def __pos__(self):
        r"""
        Return ``self``.

        OUTPUT: a :class:`ClassFunction`

        EXAMPLES::

            sage: chi = ClassFunction(SymmetricGroup(4), [3, 1, -1, 0, -1])
            sage: +chi
            Character of Symmetric group of order 4! as a permutation group
            sage: _.values()
            [3, 1, -1, 0, -1]
            sage: chi.__pos__() == +chi
            True
        """
        return ClassFunction(self._group, self._gap_classfunction)

    def __neg__(self):
        r"""
        Return the additive inverse of ``self``.

        OUTPUT: a :class:`ClassFunction`

        EXAMPLES::

            sage: chi = ClassFunction(SymmetricGroup(4), [3, 1, -1, 0, -1])
            sage: -chi
            Character of Symmetric group of order 4! as a permutation group
            sage: _.values()
            [-3, -1, 1, 0, 1]
            sage: chi.__neg__() == -chi
            True
        """
        return ClassFunction(self._group, -self._gap_classfunction)

    def __pow__(self, other):
        r"""
        Return the product of ``self`` with itself ``other`` times.

        EXAMPLES::

            sage: chi = ClassFunction(SymmetricGroup(4), [3, 1, -1, 0, -1])
            sage: p = chi**3
            sage: p
            Character of Symmetric group of order 4! as a permutation group
            sage: p.values()
            [27, 1, -1, 0, -1]
        """
        if not isinstance(other, (int, Integer)):
            raise NotImplementedError
        return ClassFunction(self._group, self._gap_classfunction ** other)

    def symmetric_power(self, n):
        r"""
        Return the symmetrized product of ``self`` with itself ``n`` times.

        INPUT:

        - ``n`` -- positive integer

        OUTPUT: the ``n``-th symmetrized power of ``self`` as a :class:`ClassFunction`

        EXAMPLES::

            sage: chi = ClassFunction(SymmetricGroup(4), [3, 1, -1, 0, -1])
            sage: p = chi.symmetric_power(3)
            sage: p
            Character of Symmetric group of order 4! as a permutation group
            sage: p.values()
            [10, 2, -2, 1, 0]
        """
        n = Integer(n)
        tbl = self._gap_classfunction.UnderlyingCharacterTable()
        return ClassFunction(self._group, tbl.SymmetricParts([self], n)[0])

    def exterior_power(self, n):
        r"""
        Return the antisymmetrized product of ``self`` with itself ``n`` times.

        INPUT:

        - ``n`` -- positive integer

        OUTPUT: the ``n``-th antisymmetrized power of ``self`` as a :class:`ClassFunction`

        EXAMPLES::

            sage: chi = ClassFunction(SymmetricGroup(4), [3, 1, -1, 0, -1])
            sage: p = chi.exterior_power(3)   # the highest antisymmetric power for a 3-d character
            sage: p
            Character of Symmetric group of order 4! as a permutation group
            sage: p.values()
            [1, -1, 1, 1, -1]
            sage: p == chi.determinant_character()
            True
        """
        n = Integer(n)
        tbl = self._gap_classfunction.UnderlyingCharacterTable()
        return ClassFunction(self._group, tbl.AntiSymmetricParts([self], n)[0])

    def scalar_product(self, other):
        r"""
        Return the scalar product of ``self`` with ``other``.

        EXAMPLES::

            sage: S4 = SymmetricGroup(4)
            sage: irr = S4.irreducible_characters()
            sage: [[x.scalar_product(y) for x in irr] for y in irr]
            [[1, 0, 0, 0, 0],
             [0, 1, 0, 0, 0],
             [0, 0, 1, 0, 0],
             [0, 0, 0, 1, 0],
             [0, 0, 0, 0, 1]]
        """
        return self._gap_classfunction.ScalarProduct(other).sage()

    def is_irreducible(self) -> bool:
        r"""
        Return ``True`` if ``self`` cannot be written as the sum of two nonzero
        characters of ``self``.

        EXAMPLES::

            sage: S4 = SymmetricGroup(4)
            sage: irr = S4.irreducible_characters()
            sage: [x.is_irreducible() for x in irr]
            [True, True, True, True, True]
        """
        return self._gap_classfunction.IsIrreducible().sage()

    def degree(self):
        r"""
        Return the degree of the character ``self``.

        EXAMPLES::

            sage: S5 = SymmetricGroup(5)
            sage: irr = S5.irreducible_characters()
            sage: [x.degree() for x in irr]
            [1, 4, 5, 6, 5, 4, 1]
        """
        return self._gap_classfunction.DegreeOfCharacter().sage()

    def irreducible_constituents(self) -> tuple:
        r"""
        Return a list of the characters that appear in the decomposition
        of ``self``.

        EXAMPLES::

            sage: S5 = SymmetricGroup(5)
            sage: chi = ClassFunction(S5, [22, -8, 2, 1, 1, 2, -3])
            sage: irr = chi.irreducible_constituents(); irr
            (Character of Symmetric group of order 5! as a permutation group,
             Character of Symmetric group of order 5! as a permutation group)
            sage: list(map(list, irr))
            [[4, -2, 0, 1, 1, 0, -1], [5, -1, 1, -1, -1, 1, 0]]

            sage: G = GL(2,3)
            sage: chi = ClassFunction(G, [-1, -1, -1, -1, -1, -1, -1, -1])
            sage: chi.irreducible_constituents()
            (Character of General Linear Group of degree 2 over Finite Field of size 3,)
            sage: chi = ClassFunction(G, [1, 1, 1, 1, 1, 1, 1, 1])
            sage: chi.irreducible_constituents()
            (Character of General Linear Group of degree 2 over Finite Field of size 3,)
            sage: chi = ClassFunction(G, [2, 2, 2, 2, 2, 2, 2, 2])
            sage: chi.irreducible_constituents()
            (Character of General Linear Group of degree 2 over Finite Field of size 3,)
            sage: chi = ClassFunction(G, [-1, -1, -1, -1, 3, -1, -1, 1])
            sage: ic = chi.irreducible_constituents(); ic
            (Character of General Linear Group of degree 2 over Finite Field of size 3,
             Character of General Linear Group of degree 2 over Finite Field of size 3)
            sage: list(map(list, ic))
            [[2, -1, 2, -1, 2, 0, 0, 0], [3, 0, 3, 0, -1, 1, 1, -1]]
        """
        L = self._gap_classfunction.ConstituentsOfCharacter()
        return tuple(ClassFunction(self._group, ell) for ell in L)

    def decompose(self) -> tuple:
        r"""
        Return a list of the characters appearing the decomposition of ``self``.

        EXAMPLES::

            sage: S5 = SymmetricGroup(5)
            sage: chi = ClassFunction(S5, [22, -8, 2, 1, 1, 2, -3])
            sage: chi.decompose()
            ((3, Character of Symmetric group of order 5! as a permutation group),
             (2, Character of Symmetric group of order 5! as a permutation group))
        """
        L = [(self.scalar_product(irr), irr)
             for irr in self.irreducible_constituents()]
        return tuple(L)

    def norm(self):
        r"""
        Return the norm of ``self``.

        EXAMPLES::

            sage: A5 = AlternatingGroup(5)
            sage: [x.norm() for x in A5.irreducible_characters()]
            [1, 1, 1, 1, 1]
        """
        return self._gap_classfunction.Norm().sage()

    def values(self) -> list:
        r"""
        Return the list of values of ``self`` on the conjugacy classes.

        EXAMPLES::

            sage: G = GL(2,3)
            sage: [x.values() for x in G.irreducible_characters()]  # random
            [[1, 1, 1, 1, 1, 1, 1, 1],
             [1, 1, 1, 1, 1, -1, -1, -1],
             [2, -1, 2, -1, 2, 0, 0, 0],
             [2, 1, -2, -1, 0, -zeta8^3 - zeta8, zeta8^3 + zeta8, 0],
             [2, 1, -2, -1, 0, zeta8^3 + zeta8, -zeta8^3 - zeta8, 0],
             [3, 0, 3, 0, -1, -1, -1, 1],
             [3, 0, 3, 0, -1, 1, 1, -1],
             [4, -1, -4, 1, 0, 0, 0, 0]]

        TESTS::

            sage: G = GL(2,3)
            sage: k = CyclotomicField(8)
            sage: zeta8 = k.gen()
            sage: v = [tuple(x.values()) for x in G.irreducible_characters()]
            sage: set(v) == {(1, 1, 1, 1, 1, 1, 1, 1), (1, 1, 1, 1, 1, -1, -1, -1), (2, -1, 2, -1, 2, 0, 0, 0), (2, 1, -2, -1, 0, -zeta8^3 - zeta8, zeta8^3 + zeta8, 0), (2, 1, -2, -1, 0, zeta8^3 + zeta8, -zeta8^3 - zeta8, 0), (3, 0, 3, 0, -1, -1, -1, 1), (3, 0, 3, 0, -1, 1, 1, -1), (4, -1, -4, 1, 0, 0, 0, 0)}
            True
        """
        return list(self)

    def central_character(self):
        r"""
        Return the central character of ``self``.

        EXAMPLES::

            sage: t = SymmetricGroup(4).trivial_character()
            sage: t.central_character().values()
            [1, 6, 3, 8, 6]
        """
        return ClassFunction(self._group, self._gap_classfunction.CentralCharacter())

    def determinant_character(self):
        r"""
        Return the determinant character of ``self``.

        EXAMPLES::

            sage: t = ClassFunction(SymmetricGroup(4), [1, -1, 1, 1, -1])
            sage: t.determinant_character().values()
            [1, -1, 1, 1, -1]
        """
        return ClassFunction(self._group, self._gap_classfunction.DeterminantOfCharacter())

    def tensor_product(self, other):
        r"""
        Return the tensor product of ``self`` and ``other``.

        EXAMPLES::

            sage: S3 = SymmetricGroup(3)
            sage: chi1, chi2, chi3 = S3.irreducible_characters()
            sage: chi1.tensor_product(chi3).values()
            [1, -1, 1]
        """
        product = libgap.Tensored([self], [other])
        return ClassFunction(self._group, product[0])

    def restrict(self, H):
        r"""
        Return the restricted character.

        INPUT:

        - ``H`` -- a subgroup of the underlying group of ``self``

        OUTPUT: a :class:`ClassFunction` of ``H`` defined by restriction

        EXAMPLES::

            sage: G = SymmetricGroup(5)
            sage: chi = ClassFunction(G, [3, -3, -1, 0, 0, -1, 3]); chi
            Character of Symmetric group of order 5! as a permutation group
            sage: H = G.subgroup([(1,2,3), (1,2), (4,5)])
            sage: chi.restrict(H)
            Character of Subgroup generated by [(1,2,3), (1,2), (4,5)] of
             (Symmetric group of order 5! as a permutation group)
            sage: chi.restrict(H).values()
            [3, -3, -3, -1, 0, 0]
        """
        try:
            gapH = H.gap()
        except AttributeError:
            from sage.libs.gap.libgap import libgap
            gapH = libgap(H)
        rest = self._gap_classfunction.RestrictedClassFunction(gapH)
        return ClassFunction(H, rest)

    def induct(self, G):
        r"""
        Return the induced character.

        INPUT:

        - ``G`` -- a supergroup of the underlying group of ``self``

        OUTPUT:

        A :class:`ClassFunction` of ``G`` defined by
        induction. Induction is the adjoint functor to restriction,
        see :meth:`restrict`.

        EXAMPLES::

            sage: G = SymmetricGroup(5)
            sage: H = G.subgroup([(1,2,3), (1,2), (4,5)])
            sage: xi = H.trivial_character(); xi
            Character of Subgroup generated by [(1,2,3), (1,2), (4,5)] of
             (Symmetric group of order 5! as a permutation group)
            sage: xi.induct(G)
            Character of Symmetric group of order 5! as a permutation group
            sage: xi.induct(G).values()
            [10, 4, 2, 1, 1, 0, 0]
        """
        try:
            gapG = G.gap()
        except AttributeError:
            from sage.libs.gap.libgap import libgap
            gapG = libgap(G)
        ind = self._gap_classfunction.InducedClassFunction(gapG)
        return ClassFunction(G, ind)

    def adams_operation(self, k):
        r"""
        Return the ``k``-th Adams operation on ``self``.

        Let `G` be a finite group. The `k`-th Adams operation `\Psi^k`
        is given by

        .. MATH::

            \Psi^k(\chi)(g) = \chi(g^k).

        The Adams operations turn the representation ring of `G`
        into a `\lambda`-ring.

        EXAMPLES::

            sage: G = GL(2,3)
            sage: chars = G.irreducible_characters()
            sage: [chi.adams_operation(2).values() for chi in chars]
            [[1, 1, 1, 1, 1, 1, 1, 1],
             [1, 1, 1, 1, 1, 1, 1, 1],
             [2, -1, 2, -1, 2, 2, 2, 2],
             [2, -1, 2, -1, -2, 0, 0, 2],
             [2, -1, 2, -1, -2, 0, 0, 2],
             [3, 0, 3, 0, 3, -1, -1, 3],
             [3, 0, 3, 0, 3, -1, -1, 3],
             [4, 1, 4, 1, -4, 0, 0, 4]]
            sage: chars[5].adams_operation(3).decompose()
            ((1, Character of General Linear Group of degree 2 over Finite Field of size 3),
             (1, Character of General Linear Group of degree 2 over Finite Field of size 3),
             (-1, Character of General Linear Group of degree 2 over Finite Field of size 3),
             (1, Character of General Linear Group of degree 2 over Finite Field of size 3))

        REFERENCES:

        - :wikipedia:`Adams_operation`
        """
        reprs = self._group.conjugacy_classes_representatives()
        return ClassFunction(self._group, [self(x**k) for x in reprs])


ClassFunction_libgap = deprecated_function_alias(41579, ClassFunction)



r"""
Misc matrix algorithms using MPFR
"""

from cysignals.signals cimport sig_check

cimport sage.rings.abc

from sage.libs.mpfr cimport *
from sage.rings.real_mpfr cimport RealNumber

from sage.matrix.matrix0 cimport Matrix


def hadamard_row_bound_mpfr(Matrix A):
    r"""
    Given a matrix `A` with entries that coerce to ``RR``, compute the row
    Hadamard bound on the determinant.

    INPUT:

    - ``A`` -- a matrix over ``RR``

    OUTPUT:

    integer -- an integer n such that the absolute value of the
    determinant of this matrix is at most `10^n`.

    EXAMPLES:

    We create a very large matrix, compute the row Hadamard bound,
    and also compute the row Hadamard bound of the transpose, which
    happens to be sharp. ::

        sage: a = matrix(ZZ, 2, [2^10000, 3^10000, 2^50, 3^19292])
        sage: from sage.matrix.misc_mpfr import hadamard_row_bound_mpfr
        sage: hadamard_row_bound_mpfr(a.change_ring(RR))
        13976
        sage: len(str(a.det()))
        12215
        sage: hadamard_row_bound_mpfr(a.transpose().change_ring(RR))
        12215

    Note that in the above example using RDF would overflow::

        sage: b = a.change_ring(RDF)
        sage: b._hadamard_row_bound()
        Traceback (most recent call last):
        ...
        OverflowError: cannot convert float infinity to integer
    """
    if not isinstance(A.base_ring(), sage.rings.abc.RealField):
        raise TypeError("A must have base field an mpfr real field.")

    cdef RealNumber a, b
    cdef mpfr_t s, d, pr
    cdef Py_ssize_t i, j

    mpfr_init(s)
    mpfr_init(d)
    mpfr_init(pr)
    mpfr_set_si(d, 0, MPFR_RNDU)

    for i in range(A._nrows):
        mpfr_set_si(s, 0, MPFR_RNDU)
        for j in range(A._ncols):
            sig_check()
            a = A.get_unsafe(i, j)
            mpfr_mul(pr, a.value, a.value, MPFR_RNDU)
            mpfr_add(s, s, pr, MPFR_RNDU)
        mpfr_log10(s, s, MPFR_RNDU)
        mpfr_add(d, d, s, MPFR_RNDU)
    b = a._new()
    mpfr_set(b.value, d, MPFR_RNDU)
    b /= 2
    mpfr_clear(s)
    mpfr_clear(d)
    mpfr_clear(pr)
    return b.ceil()



# distutils: libraries = flint
# distutils: depends = flint/nmod_poly_mat.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void nmod_poly_mat_init(nmod_poly_mat_t mat, slong rows, slong cols, mp_limb_t n) noexcept
    void nmod_poly_mat_init_set(nmod_poly_mat_t mat, const nmod_poly_mat_t src) noexcept
    void nmod_poly_mat_clear(nmod_poly_mat_t mat) noexcept
    void nmod_poly_mat_set_trunc(nmod_poly_mat_t res, const nmod_poly_mat_t pmat, long len) noexcept
    void nmod_poly_mat_truncate(nmod_poly_mat_t pmat, long len) noexcept
    void nmod_poly_mat_shift_left(nmod_poly_mat_t res, const nmod_poly_mat_t pmat, slong k) noexcept
    void nmod_poly_mat_shift_right(nmod_poly_mat_t res, const nmod_poly_mat_t pmat, slong k) noexcept
    slong nmod_poly_mat_nrows(const nmod_poly_mat_t mat) noexcept
    slong nmod_poly_mat_ncols(const nmod_poly_mat_t mat) noexcept
    mp_limb_t nmod_poly_mat_modulus(const nmod_poly_mat_t mat) noexcept
    nmod_poly_struct * nmod_poly_mat_entry(const nmod_poly_mat_t mat, slong i, slong j) noexcept
    void nmod_poly_mat_set(nmod_poly_mat_t mat1, const nmod_poly_mat_t mat2) noexcept
    void nmod_poly_mat_set_nmod_mat(nmod_poly_mat_t pmat, const nmod_mat_t cmat) noexcept
    void nmod_poly_mat_swap(nmod_poly_mat_t mat1, nmod_poly_mat_t mat2) noexcept
    void nmod_poly_mat_swap_entrywise(nmod_poly_mat_t mat1, nmod_poly_mat_t mat2) noexcept
    void nmod_poly_mat_print(const nmod_poly_mat_t mat, const char * x) noexcept
    void nmod_poly_mat_randtest(nmod_poly_mat_t mat, flint_rand_t state, slong len) noexcept
    void nmod_poly_mat_randtest_sparse(nmod_poly_mat_t A, flint_rand_t state, slong len, float density) noexcept
    void nmod_poly_mat_zero(nmod_poly_mat_t mat) noexcept
    void nmod_poly_mat_one(nmod_poly_mat_t mat) noexcept
    bint nmod_poly_mat_equal(const nmod_poly_mat_t mat1, const nmod_poly_mat_t mat2) noexcept
    bint nmod_poly_mat_equal_nmod_mat(const nmod_poly_mat_t pmat, const nmod_mat_t cmat) noexcept
    bint nmod_poly_mat_is_zero(const nmod_poly_mat_t mat) noexcept
    bint nmod_poly_mat_is_one(const nmod_poly_mat_t mat) noexcept
    bint nmod_poly_mat_is_empty(const nmod_poly_mat_t mat) noexcept
    bint nmod_poly_mat_is_square(const nmod_poly_mat_t mat) noexcept
    void nmod_poly_mat_get_coeff_mat(nmod_mat_t coeff, const nmod_poly_mat_t pmat, slong deg) noexcept
    void nmod_poly_mat_set_coeff_mat(nmod_poly_mat_t pmat, const nmod_mat_t coeff, slong deg) noexcept
    slong nmod_poly_mat_max_length(const nmod_poly_mat_t A) noexcept
    slong nmod_poly_mat_degree(const nmod_poly_mat_t pmat) noexcept
    void nmod_poly_mat_evaluate_nmod(nmod_mat_t B, const nmod_poly_mat_t A, mp_limb_t x) noexcept
    void nmod_poly_mat_scalar_mul_nmod_poly(nmod_poly_mat_t B, const nmod_poly_mat_t A, const nmod_poly_t c) noexcept
    void nmod_poly_mat_scalar_mul_nmod(nmod_poly_mat_t B, const nmod_poly_mat_t A, mp_limb_t c) noexcept
    void nmod_poly_mat_add(nmod_poly_mat_t C, const nmod_poly_mat_t A, const nmod_poly_mat_t B) noexcept
    void nmod_poly_mat_sub(nmod_poly_mat_t C, const nmod_poly_mat_t A, const nmod_poly_mat_t B) noexcept
    void nmod_poly_mat_neg(nmod_poly_mat_t B, const nmod_poly_mat_t A) noexcept
    void nmod_poly_mat_mul(nmod_poly_mat_t C, const nmod_poly_mat_t A, const nmod_poly_mat_t B) noexcept
    void nmod_poly_mat_mul_classical(nmod_poly_mat_t C, const nmod_poly_mat_t A, const nmod_poly_mat_t B) noexcept
    void nmod_poly_mat_mul_KS(nmod_poly_mat_t C, const nmod_poly_mat_t A, const nmod_poly_mat_t B) noexcept
    void nmod_poly_mat_mul_interpolate(nmod_poly_mat_t C, const nmod_poly_mat_t A, const nmod_poly_mat_t B) noexcept
    void nmod_poly_mat_sqr(nmod_poly_mat_t B, const nmod_poly_mat_t A) noexcept
    void nmod_poly_mat_sqr_classical(nmod_poly_mat_t B, const nmod_poly_mat_t A) noexcept
    void nmod_poly_mat_sqr_KS(nmod_poly_mat_t B, const nmod_poly_mat_t A) noexcept
    void nmod_poly_mat_sqr_interpolate(nmod_poly_mat_t B, const nmod_poly_mat_t A) noexcept
    void nmod_poly_mat_pow(nmod_poly_mat_t B, const nmod_poly_mat_t A, ulong exp) noexcept
    slong nmod_poly_mat_find_pivot_any(const nmod_poly_mat_t mat, slong start_row, slong end_row, slong c) noexcept
    slong nmod_poly_mat_find_pivot_partial(const nmod_poly_mat_t mat, slong start_row, slong end_row, slong c) noexcept
    slong nmod_poly_mat_fflu(nmod_poly_mat_t B, nmod_poly_t den, slong * perm, const nmod_poly_mat_t A, int rank_check) noexcept
    slong nmod_poly_mat_rref(nmod_poly_mat_t B, nmod_poly_t den, const nmod_poly_mat_t A) noexcept
    void nmod_poly_mat_trace(nmod_poly_t trace, const nmod_poly_mat_t mat) noexcept
    void nmod_poly_mat_det(nmod_poly_t det, const nmod_poly_mat_t A) noexcept
    void nmod_poly_mat_det_fflu(nmod_poly_t det, const nmod_poly_mat_t A) noexcept
    void nmod_poly_mat_det_interpolate(nmod_poly_t det, const nmod_poly_mat_t A) noexcept
    slong nmod_poly_mat_rank(const nmod_poly_mat_t A) noexcept
    int nmod_poly_mat_inv(nmod_poly_mat_t Ainv, nmod_poly_t den, const nmod_poly_mat_t A) noexcept
    slong nmod_poly_mat_nullspace(nmod_poly_mat_t res, const nmod_poly_mat_t mat) noexcept
    int nmod_poly_mat_solve(nmod_poly_mat_t X, nmod_poly_t den, const nmod_poly_mat_t A, const nmod_poly_mat_t B) noexcept
    int nmod_poly_mat_solve_fflu(nmod_poly_mat_t X, nmod_poly_t den, const nmod_poly_mat_t A, const nmod_poly_mat_t B) noexcept
    void nmod_poly_mat_solve_fflu_precomp(nmod_poly_mat_t X, const slong * perm, const nmod_poly_mat_t FFLU, const nmod_poly_mat_t B) noexcept



r"""
The Tachyon Ray Tracer

AUTHOR:

- John E. Stone

This documentation, which was written by John Stone, describes how to
create scene files.

At the present time, scene description files are very simple. The parser
can’t handle multiple file scene descriptions, although they may be
added in the future. Most of the objects and their scene description are
closely related to the RAY API. *(See the API docs for additional info.)*

Basic Scene Requirements
------------------------

Unlike some other ray tracers out there, RAY requires that you specify
most of the scene parameters in the scene description file itself. If
users would rather specify some of these parameters at the command line,
then I may add that feature in the future. A scene description file
contains keywords, and values associated or grouped with a keyword. All
keywords can be in caps, lower case, or mixed case for the convenience
of the user. File names and texture names are normally case-sensitive,
although the behavior for file names is operating system-dependent. All
values are either character strings, or floating point numbers. In some
cases, the presence of one keyword will require additional keyword /
value pairs.

At the moment there are several keywords with values, that must appear
in every scene description file. Every scene description file must begin
with the ``BEGIN_SCENE`` keyword, and end with the ``END_SCENE``
keyword. All definitions and declarations of any kind must be inside the
``BEGIN_SCENE``, ``END_SCENE`` pair. The ``RESOLUTION`` keyword is
followed by an x resolution and a y resolution in terms of pixels on
each axis. There are currently no limits placed on the resolution of an
output image other than the computer’s available memory and reasonable
execution time. An example of a simple scene description skeleton is
show below:

::

   BEGIN_SCENE
     RESOLUTION 1024 1024
   ...
   ...  Camera definition..
   ...
   ...  Other objects, etc..
   ...

   END_SCENE

Camera and viewing parameters
-----------------------------

One of the most important parts of any scene, is the camera position and
orientation. Having a good angle on a scene can make the difference
between an average looking scene and a strikingly interesting one. There
may be multiple camera definitions in a scene file, but the last camera
definition overrides all previous definitions. There are several
parameters that control the camera in , ``PROJECTION``, ``ZOOM``,
``ASPECTRATIO``, ``ANTIALIASING``, ``CENTER``, ``RAYDEPTH``,
``VIEWDIR``, and ``UPDIR``.

The first and last keywords required in the definition of a camera are
the ``CAMERA`` and ``END_CAMERA`` keywords. The ``PROJECTION`` keyword
is optional, the remaining camera keywords are required, and must be
written in the sequence they are listed in the examples in this section.

Camera projection modes
~~~~~~~~~~~~~~~~~~~~~~~

The ``PROJECTION`` keyword must be followed by one of the supported
camera projection mode identifiers ``PERSPECTIVE``, ``PERSPECTIVE_DOF``,
``ORTHOGRAPHIC``, or ``FISHEYE``. The ``FISHEYE`` projection mode
requires two extra parameters ``FOCALLENGTH`` and ``APERTURE`` which
precede the regular camera options.

::

   Camera
     projection perspective_dof
     focallength 0.75
     aperture 0.02
     Zoom 0.666667
     Aspectratio 1.000000
     Antialiasing 128
     Raydepth 30
     Center  0.000000 0.000000 -2.000000
     Viewdir -0.000000 -0.000000 2.000000
     Updir   0.000000 1.000000 -0.000000
   End_Camera

Common camera parameters
~~~~~~~~~~~~~~~~~~~~~~~~

The ``ZOOM`` parameter controls the camera in a way similar to a
telephoto lens on a 35mm camera. A zoom value of 1.0 is standard, with a
90 degree field of view. By changing the zoom factor to 2.0, the
relative size of any feature in the frame is twice as big, while the
field of view is decreased slightly. The zoom effect is implemented as a
scaling factor on the height and width of the image plane relative to
the world.

The ``ASPECTRATIO`` parameter controls the aspect ratio of the resulting
image. By using the aspect ratio parameter, one can produce images which
look correct on any screen. Aspect ratio alters the relative width of
the image plane, while keeping the height of the image plane constant.
In general, most workstation displays have an aspect ratio of 1.0. To
see what aspect ratio your display has, you can render a simple sphere,
at a resolution of 512x512 and measure the ratio of its width to its
height.

The ``ANTIALIASING`` parameter controls the maximum level of
supersampling used to obtain higher image quality. The parameter given
sets the number of additional rays to trace per-pixel to attain higher
image quality.

The ``RAYDEPTH`` parameter tells RAY what the maximum level of
reflections, refraction, or in general the maximum recursion depth to
trace rays to. A value between 4 and 12 is usually good. A value of 1
will disable rendering of reflective or transmissive objects (they’ll be
black).

The remaining three camera parameters are the most important, because
they define the coordinate system of the camera, and its position in the
scene. The ``CENTER`` parameter is an X, Y, Z coordinate defining the
center of the camera *(also known as the Center of Projection)*. Once
you have determined where the camera will be placed in the scene, you
need to tell RAY what the camera should be looking at. The ``VIEWDIR``
parameter is a vector indicating the direction the camera is facing. It
may be useful for me to add a "Look At" type keyword in the future to
make camera aiming easier. If people want or need the "Look At" style
camera, let me know. The last parameter needed to completely define a
camera is the "up" direction. The ``UPDIR`` parameter is a vector which
points in the direction of the "sky". I wrote the camera so that
``VIEWDIR`` and ``UPDIR`` don’t have to be perpendicular, and there
shouldn’t be a need for a "right" vector although some other ray tracers
require it. Here’s a snippet of a camera definition:

::

   CAMERA
     ZOOM 1.0
     ASPECTRATIO 1.0
     ANTIALIASING 0
     RAYDEPTH 12
     CENTER 0.0 0.0 2.0
     VIEWDIR 0 0 -1
     UPDIR 0 1 0
   END_CAMERA

Viewing frustum
~~~~~~~~~~~~~~~

An optional ``FRUSTUM`` parameter provides a means for rendering
sub-images in a larger frame, and correct stereoscopic images. The
``FRUSTUM`` keyword must be followed by four floating parameters, which
indicate the top, bottom, left and right coordinates of the image plane
in eye coordinates. When the projection mode is set to ``FISHEYE``, the
frustum parameters correspond to spherical coordinates specified in
radians.

::

   CAMERA
     ZOOM  1.0
     ASPECTRATIO 1.0
     ANTIALIASING 0
     RAYDEPTH  4
     CENTER    0.0 0.0 -6.0
     VIEWDIR   0.0 0.0 1.0
     UPDIR     0.0 1.0 0.0
     FRUSTUM   -0.5 0.5 -0.5 0.5
   END_CAMERA

Including Files
---------------

The ``INCLUDE`` keyword is used anywhere after the camera description,
and is immediately followed by a valid filename, for a file containing
additional scene description information. The included file is opened,
and processing continues as if it were part of the current file, until
the end of the included file is reached. Parsing of the current file
continues from where it left off prior to the included file.

Scene File Comments
-------------------

The ``#`` keyword is used anywhere after the camera
description, and will cause RAY to ignore all characters from the
``#`` to the end of the input line. The ``#``
character must be surrounded by whitespace in order to be recognized. A
sequence such as ``###`` will not be recognized as a comment.

Lights
------

The most frequently used type of lights provided by RAY are positional
point light sources. The lights are actually small spheres, which are
visible. A point light is composed of three pieces of information, a
center, a radius (since its a sphere), and a color. To define a light,
simply write the ``LIGHT`` keyword, followed by its ``CENTER`` (a X, Y,
Z coordinate), its ``RAD`` (radius, a scalar), and its ``COLOR`` (a Red
Green Blue triple). The radius parameter will accept any value of 0.0 or
greater. Lights of radius 0.0 will not be directly visible in the
rendered scene, but contribute light to the scene normally. For a light,
the color values range from 0.0 to 1.0, any values outside this range
may yield unpredictable results. A simple light definition looks like
this:

::

     LIGHT CENTER 4.0 3.0 2.0
           RAD    0.2
           COLOR  0.5 0.5 0.5

This light would be gray colored if seen directly, and would be 50%
intensity in each RGB color component.

RAY supports simple directional lighting, commonly used in CAD and
scientific visualization programs for its performance advantages over
positional lights. Directional lights cannot be seen directly in scenes
rendered by , only their illumination contributes to the final image.

::

   DIRECTIONAL_LIGHT
     DIRECTION 0.0 -1.0 0.0
     COLOR   1.0 0.0  0.0

RAY supports spotlights, which are described very similarly to a point
light, but they are attenuated by angle from the direction vector, based
on a “falloff start” angle and “falloff end”angle. Between the starting
and ending angles, the illumination is attenuated linearly. The syntax
for a spotlight description in a scene file is as follows.

::

   SPOTLIGHT
     CENTER  0.0 3.0  17.0
     RAD     0.2
     DIRECTION 0.0 -1.0 0.0
       FALLOFF_START 20.0
       FALLOFF_END   45.0
     COLOR   1.0 0.0  0.0

The lighting system implemented by RAY provides various levels of
distance-based lighting attenuation. By default, a light is not
attenuated by distance. If the *attenuation* keywords is present
immediately prior to the light’s color, RAY will accept coefficients
which are used to calculate distance-based attenuation, which is applied
the light by multiplying with the resulting value. The attenuation
factor is calculated from the equation

.. math:: 1/(K_c + K_l d + k_q d^2)

This attenuation equation should be familiar to some as it is the same
lighting attenuation equation used by OpenGL. The constant, linear, and
quadratic terms are specified in a scene file as shown in the following
example.

::

   LIGHT
     CENTER  -5.0 0.0 10.0
     RAD     1.0
     ATTENUATION CONSTANT 1.0 LINEAR 0.2 QUADRATIC 0.05
     COLOR   1.0 0.0 0.0

Atmospheric effects
-------------------

RAY currently only implements one atmospheric effect, simple
distance-based fog.

Fog
~~~

RAY provides a simple distance-based fog effect intended to provide
functionality similar to that found in OpenGL, for compatibility with
software that requires an OpenGL-like fog implementation. Much like
OpenGL, RAY provides linear, exponential, and exponential-squared fog.

::

     FOG
       LINEAR START 0.0  END 50.0  DENSITY 1.0  COLOR 1.0 1.0 1.0

::

     FOG
       EXP START 0.0  END 50.0  DENSITY 1.0  COLOR 1.0 1.0 1.0

::

     FOG
       EXP2 START 0.0  END 50.0  DENSITY 1.0  COLOR 1.0 1.0 1.0

Objects
-------

Spheres
~~~~~~~

Spheres are the simplest object supported by RAY and they are also the
fastest object to render. Spheres are defined as one would expect, with
a ``CENTER``, ``RAD`` (radius), and a texture. The texture may be
defined along with the object as discussed earlier, or it may be
declared and assigned a name. Here’s a sphere definition using a
previously defined "NitrogenAtom" texture:

::

    SPHERE  CENTER 26.4 27.4 -2.4   RAD 1.0   NitrogenAtom

A sphere with an inline texture definition is declared like this:

::

    Sphere center 1.0 0.0 10.0
              Rad 1.0
           Texture  Ambient 0.2  Diffuse 0.8  Specular 0.0  Opacity 1.0
                    Color   1.0 0.0 0.5
                    TexFunc 0

Notice that in this example I used mixed case for the keywords, this is
allowable... Review the section on textures if the texture definitions
are confusing.

Triangles
~~~~~~~~~

Triangles are also fairly simple objects, constructed by listing the
three vertices of the triangle, and its texture. The order of the
vertices isn’t important, the triangle object is "double sided", so the
surface normal is always pointing back in the direction of the incident
ray. The triangle vertices are listed as ``V1``, ``V2``, and ``V3`` each
one is an X, Y, Z coordinate. An example of a triangle is shown below:

::

   TRI
     V0  0.0 -4.0 12.0
     V1  4.0 -4.0 8.0
     V2 -4.0 -4.0 8.0
     TEXTURE
       AMBIENT  0.1 DIFFUSE  0.2 SPECULAR 0.7 OPACITY 1.0
       COLOR 1.0 1.0 1.0
       TEXFUNC 0

Smoothed Triangles
~~~~~~~~~~~~~~~~~~

Smoothed triangles are just like regular triangles, except that the
surface normal for each of the three vertices is used to determine the
surface normal across the triangle by linear interpolation. Smoothed
triangles yield curved looking objects and have nice reflections.

::

   STRI
     V0 1.4   0.0   2.4
     V1 1.35 -0.37  2.4
     V2 1.36 -0.32  2.45
     N0 -0.9 -0.0  -0.4
     N1 -0.8  0.23 -0.4
     N2 -0.9  0.27 -0.15
     TEXTURE
       AMBIENT  0.1 DIFFUSE  0.2 SPECULAR 0.7 OPACITY 1.0
       COLOR 1.0 1.0 1.0
       TEXFUNC 0

Infinite Planes
~~~~~~~~~~~~~~~

Useful for things like desert floors, backgrounds, skies etc, the
infinite plane is pretty easy to use. An infinite plane only consists of
two pieces of information, the ``CENTER`` of the plane, and a ``NORMAL``
to the plane. The center of the plane is just any point on the plane
such that the point combined with the surface normal define the equation
for the plane. As with triangles, planes are double sided. Here is an
example of an infinite plane:

::

   PLANE
     CENTER 0.0 -5.0 0.0
     NORMAL 0.0  1.0 0.0
     TEXTURE
       AMBIENT  0.1 DIFFUSE  0.9 SPECULAR 0.0  OPACITY 1.0
       COLOR  1.0 1.0 1.0
       TEXFUNC  1
         CENTER 0.0 -5.0 0.0
         ROTATE 0. 0.0 0.0
         SCALE  1.0 1.0 1.0

Rings
~~~~~

Rings are a simple object, they are really a not-so-infinite plane.
Rings are simply an infinite plane cut into a washer shaped ring,
infinitely thing just like a plane. A ring only requires two more pieces
of information than an infinite plane does, an inner and outer radius.
Here’s an example of a ring:

::

     Ring
       Center 1.0 1.0 1.0
       Normal 0.0 1.0 0.0
       Inner  1.0
       Outer  5.0
       MyNewRedTexture

Infinite Cylinders
~~~~~~~~~~~~~~~~~~

Infinite cylinders are quite simple. They are defined by a center, an
axis, and a radius. An example of an infinite cylinder is:

::

     Cylinder
       Center 0.0 0.0 0.0
       Axis   0.0 1.0 0.0
       Rad    1.0
       SomeRandomTexture

Finite Cylinders
~~~~~~~~~~~~~~~~

Finite cylinders are almost the same as infinite ones, but the center
and length of the axis determine the extents of the cylinder. The finite
cylinder is also really a shell, it doesn’t have any caps. If you need
to close off the ends of the cylinder, use two ring objects, with the
inner radius set to 0.0 and the normal set to be the axis of the
cylinder. Finite cylinders are built this way to enhance speed.

::

     FCylinder
       Center 0.0 0.0 0.0
       Axis   0.0 9.0 0.0
       Rad    1.0
       SomeRandomTexture

This defines a finite cylinder with radius 1.0, going from 0.0 0.0 0.0,
to 0.0 9.0 0.0 along the Y axis. The main difference between an infinite
cylinder and a finite cylinder is in the interpretation of the ``AXIS``
parameter. In the case of the infinite cylinder, the length of the axis
vector is ignored. In the case of the finite cylinder, the axis
parameter is used to determine the length of the overall cylinder.

Axis Aligned Boxes
~~~~~~~~~~~~~~~~~~

Axis aligned boxes are fast, but of limited usefulness. As such, I’m not
going to waste much time explaining ’em. An axis aligned box is defined
by a ``MIN`` point, and a ``MAX`` point. The volume between the min and
max points is the box. Here’s a simple box:

::

     BOX
       MIN -1.0 -1.0 -1.0
       MAX  1.0  1.0  1.0
       Boxtexture1

Fractal Landscapes
~~~~~~~~~~~~~~~~~~

Currently fractal landscapes are a built-in function. In the near future
I’ll allow the user to load an image map for use as a heightfield.
Fractal landscapes are currently forced to be axis aligned. Any
suggestion on how to make them more appealing to users is welcome. A
fractal landscape is defined by its "resolution" which is the number of
grid points along each axis, and by its scale and center. The "scale" is
how large the landscape is along the X, and Y axes in world coordinates.
Here’s a simple landscape:

::

   SCAPE
     RES 30 30
     SCALE 80.0 80.0
     CENTER 0.0 -4.0 20.0
     TEXTURE
       AMBIENT 0.1 DIFFUSE 0.9 SPECULAR 0.0 OPACITY 1.0
       COLOR 1.0 1.0 1.0
       TEXFUNC 0

The landscape shown above generates a square landscape made of 1,800
triangles. When time permits, the heightfield code will be rewritten to
be more general and to increase rendering speed.

Arbitrary Quadric Surfaces
~~~~~~~~~~~~~~~~~~~~~~~~~~

Docs soon. I need to add these into the parser, must have forgotten
before ;-)

Volume Rendered Scalar Voxels
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

These are a little trickier than the average object :-) These are likely
to change substantially in the very near future so I’m not going to get
too detailed yet. A volume rendered data set is described by its axis
aligned bounding box, and its resolution along each axis. The final
parameter is the voxel data file. If you are seriously interested in
messing with these, get hold of me and I’ll give you more info. Here’s a
quick example:

::

   SCALARVOL
     MIN -1.0 -1.0 -0.4
     MAX  1.0  1.0  0.4
     DIM 256 256 100
     FILE /cfs/johns/vol/engine.256x256x110
     TEXTURE
           AMBIENT 1.0 DIFFUSE 0.0 SPECULAR 0.0 OPACITY 8.1
           COLOR 1.0 1.0 1.0
           TEXFUNC 0

Texture and Color
-----------------

Simple Texture Characteristics
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The surface textures applied to an object drastically alter its overall
appearance, making textures and color one of the most important topics
in this manual. As with many other renderers, textures can be declared
and associated with a name so that they may be used over and over again
in a scene definition with less typing. If a texture is only need once,
or it is unique to a particular object in the scene, then it may be
declared along with the object it is applied to, and does not need a
name.

The simplest texture definition is a solid color with no image mapping
or procedural texture mapping. A solid color texture is defined by the
``AMBIENT``, ``DIFFUSE``, ``SPECULAR``, ``OPACITY`` and ``COLOR``
parameters. The ``AMBIENT`` parameter defines the ambient lighting
coefficient to be used when shading the object. Similarly, the
``DIFFUSE`` parameter is the relative contribution of the diffuse
shading to the surface appearance. The ``SPECULAR`` parameter is the
contribution from perfectly reflected rays, as if on a mirrored surface.
``OPACITY`` defines how transparent a surface is. An ``OPACITY`` value
of 0.0 renders the object completely invisible. An ``OPACITY`` value of
1.0 makes the object completely solid, and non-transmissive. In general,
the values for the ambient, diffuse, and specular parameters should add
up to 1.0, if they don’t then pixels may be over or underexposed quite
easily. These parameters function in a manner similar to that of other
ray tracers. The ``COLOR`` parameter is an RGB triple with each value
ranging from 0.0 to 1.0 inclusive. If the RGB values stray from 0.0 to
1.0, results are undefined. In the case of solid textures, a final
parameter, ``TEXFUNC`` is set to zero (integer).

Texture Declaration and Aliasing
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

To define a simple texture for use on several objects in a scene, the
``TEXDEF`` keyword is used. The ``TEXDEF`` keyword is followed by a case
sensitive texture name, which will subsequently be used while defining
objects. If many objects in a scene use the same texture through texture
definition, a significant amount of memory may be saved since only one
copy of the texture is present in memory, and its shared by all of the
objects. Here is an example of a solid texture definition:

::

    TEXDEF MyNewRedTexture
       AMBIENT 0.1 DIFFUSE 0.9 SPECULAR 0.0 OPACITY 1.0
       COLOR 1.0 0.0 0.0  TEXFUNC 0

When this texture is used in an object definition, it is referenced only
by name. Be careful not to use one of the other keywords as a defined
texture, this will probably cause the parser to explode, as I don’t
check for use of keywords as texture names.

When a texture is declared within an object definition, it appears in an
identical format to the ``TEXDEF`` declaration, but the ``TEXTURE``
keyword is used instead of ``TEXDEF``. If it is useful to have several
names for the same texture (when you are too lazy to actually finish
defining different variations of a wood texture for example, and just
want to be approximately correct for example) aliases can be constructed
using the ``TEXALIAS`` keyword, along with the alias name, and the
original name. An example of a texture alias is:

::

     TEXALIAS MyNewestRedTexture  MyNewRedTexture

This line would alias MyNewestRedTexture to be the same thing as the
previously declared MyNewRedTexture. Note that the source texture must
be declared before any aliases that use it.

Image Maps and Procedural Textures
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Image maps and procedural textures very useful in making realistic
looking scenes. A good image map can do as much for the realism of a
wooden table as any amount of sophisticated geometry or lighting. Image
maps are made by wrapping an image on to an object in one of three ways,
a spherical map, a cylindrical map, and a planar map. Procedural
textures are used in a way similar to the image maps, but they are on
the fly and do not use much memory compared to the image maps. The main
disadvantage of the procedural maps is that they must be hard-coded into
RAY when it is compiled.

The syntax used for all texture maps is fairly simple to learn. The
biggest problem with the way that the parser is written now is that the
different mappings are selected by an integer, which is not very user
friendly. I expect to rewrite this section of the parser sometime in the
near future to alleviate this problem. When I rewrite the parser, I may
also end up altering the parameters that are used to describe a texture
map, and some of them may become optional rather than required.

.. container:: center

   +---------------------------+-----------------------------------------+
   | Texture Mapping Functions |                                         |
   +===========================+=========================================+
   | Value for TEXFUNC         | Mapping and Texture Description         |
   +---------------------------+-----------------------------------------+
   | 0                         | No special texture, plain shading       |
   +---------------------------+-----------------------------------------+
   | 1                         | 3D checkerboard function, like a        |
   |                           | Rubik’s cube                            |
   +---------------------------+-----------------------------------------+
   | 2                         | Grit Texture, randomized surface color  |
   +---------------------------+-----------------------------------------+
   | 3                         | 3D marble texture, uses object’s base   |
   |                           | color                                   |
   +---------------------------+-----------------------------------------+
   | 4                         | 3D wood texture, light and dark brown,  |
   |                           | not very good yet                       |
   +---------------------------+-----------------------------------------+
   | 5                         | 3D gradient noise function (can’t       |
   |                           | remember what it look like              |
   +---------------------------+-----------------------------------------+
   | 6                         | Don’t remember                          |
   +---------------------------+-----------------------------------------+
   | 7                         | Cylindrical Image Map, requires ppm     |
   |                           | filename                                |
   +---------------------------+-----------------------------------------+
   | 8                         | Spherical Image Map, requires ppm       |
   |                           | filename                                |
   +---------------------------+-----------------------------------------+
   | 9                         | Planar Image Map, requires ppm filename |
   +---------------------------+-----------------------------------------+

Here’s an example of a sphere, with a spherical image map applied to its
surface:

::

   SPHERE
     CENTER 2.0  0.0 5.0
     RAD 2.0
     TEXTURE
       AMBIENT  0.4 DIFFUSE  0.8 SPECULAR 0.0  OPACITY 1.0
       COLOR 1.0 1.0 1.0
       TEXFUNC 7 /cfs/johns/imaps/fire644.ppm
         CENTER 2.0 0.0 5.0
         ROTATE 0.0 0.0 0.0
         SCALE  2.0 -2.0 1.0

Basically, the image maps require the center, rotate and scale
parameters so that you can position the image map on the object
properly.
"""

# ****************************************************************************
#       Copyright (C) 2006 John E. Stone
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import os
import re

from sage.cpython.string import bytes_to_str
from sage.misc.pager import pager
from sage.misc.temporary_file import tmp_filename
from sage.structure.sage_object import SageObject
from sage.misc.cachefunc import cached_method


class TachyonRT(SageObject):
    r"""
    The Tachyon Ray Tracer.

    Usage:
    ``tachyon_rt(model, outfile='sage.png', verbose=1, block=True, extra_opts='')``

    INPUT:

    - ``model`` -- string that describes a 3d model in
      the Tachyon modeling format. Type ``sage.interfaces.tachyon?`` for a
      description of this format.

    - ``outfile`` -- (default: ``'sage.png'``) output filename;
      the extension of the filename determines the type. Supported types
      include:

      - ``tga`` -- 24-bit (uncompressed)

      - ``bmp`` -- 24-bit Windows BMP (uncompressed)

      - ``ppm`` -- 24-bit PPM (uncompressed)

      - ``rgb`` -- 24-bit SGI RGB (uncompressed)

      - ``png`` -- 24-bit PNG (compressed, lossless)

    - ``verbose`` -- integer; (default: 1)

      - ``0`` -- silent

      - ``1`` -- some output

      - ``2`` -- very verbose output

    - ``block`` -- boolean (default: ``True``); if ``False``, run the
      rendering command in the background

    - ``extra_opts`` -- passed directly to tachyon command
      line. Use ``tachyon_rt.usage()`` to see some of the possibilities

    OUTPUT: some text may be displayed onscreen

    - The file outfile is created.

    EXAMPLES:

    .. automethod:: __call__
    """
    def _repr_(self):
        """
        Return a brief description of this interface object (the Tachyon
        raytracer written by John Stone).

        TESTS::

            sage: from sage.interfaces.tachyon import TachyonRT
            sage: t = TachyonRT()
            sage: print(t.__repr__())
            John Stone's Tachyon Ray Tracer
        """
        return "John Stone's Tachyon Ray Tracer"

    def __call__(self, model, outfile='sage.png', verbose=1, extra_opts=''):
        """
        This executes the tachyon program, given a scene file input.

        INPUT:

        - ``model`` -- string; the tachyon model

        - ``outfile`` -- string (default: ``'sage.png'``); the filename
          to save the model to

        - ``verbose`` -- 0, 1, (default) or 2; the verbosity level

        - ``extra_opts`` -- string (default: empty string); extra
          options that will be appended to the tachyon commandline

        EXAMPLES::

            sage: from sage.interfaces.tachyon import TachyonRT
            sage: tgen = Tachyon()
            sage: tgen.texture('t1')
            sage: tgen.sphere((0,0,0),1,'t1')
            sage: tgen.str()[30:40]
            'resolution'
            sage: t = TachyonRT()
            sage: import os
            sage: t(tgen.str(), outfile=os.devnull)
            tachyon ...
            Tachyon Parallel/Multiprocessor Ray Tracer...

        TESTS::

            sage: from sage.env import SAGE_EXTCODE
            sage: filename = os.path.join(SAGE_EXTCODE, 'doctest', 'invalid', 'syntax_error.tachyon')
            sage: with open(filename, 'r') as f:
            ....:     syntax_error = f.read()
            sage: t(syntax_error, outfile=os.devnull)
            Traceback (most recent call last):
            ...
            RuntimeError: Tachyon Parallel/Multiprocessor Ray Tracer...
            ...
            Parser failed due to an input file syntax error.
            Aborting render.
        """
        if self.version() >= '0.99.2':
            # this keyword was changed in 0.99.2
            model = model.replace(
                "              focallength ",
                "              focaldist ")
        modelfile = tmp_filename(ext='.dat')
        with open(modelfile, 'w') as file:
            file.write(model)
        cmd = ['tachyon', modelfile]
        ext = outfile[-4:].lower()
        if ext == '.png':
            cmd += ['-format', 'PNG']
        elif ext == '.tga':
            cmd += ['-format', 'TARGA']
        elif ext == '.bmp':
            cmd += ['-format', 'BMP']
        elif ext == '.ppm':
            cmd += ['-format', 'PPM']
        elif ext == '.rgb':
            cmd += ['-format', 'RGB']
        cmd += ['-o', outfile]
        cmd += extra_opts.split()
        if verbose >= 2:
            cmd += ['+V']
        if verbose:
            print(' '.join(cmd))
        import subprocess
        out = bytes_to_str(subprocess.check_output(cmd))
        if verbose >= 1:
            print(out)
        if out.rstrip().endswith('Aborting render.'):
            raise RuntimeError(out)
        if outfile != os.devnull and os.stat(outfile).st_size == 0:
            raise RuntimeError('tachyon did not abort but output file is empty')

    def usage(self, use_pager=True):
        """
        Return the basic description of using the Tachyon raytracer (simply
        what is returned by running tachyon with no input).  The output is
        paged unless ``use_pager=False``.

        TESTS::

            sage: from sage.interfaces.tachyon import TachyonRT
            sage: t = TachyonRT()
            sage: t.usage(use_pager=False)
            ...
              tachyon modelfile [options]...
            <BLANKLINE>
            Model file formats supported:
              filename.dat ...
        """
        with os.popen('tachyon') as f:
            r = f.read()
        if use_pager:
            pager()(r)
        else:
            print(r)

    @cached_method
    def version(self):
        """
        Return the version of the Tachyon raytracer being used.

        TESTS::

            sage: tachyon_rt.version()  # random
            0.98.9
            sage: tachyon_rt.version() >= '0.98.9'
            True
        """
        with os.popen('tachyon') as f:
            r = f.readline()
        res = re.search(r"Version ([\d.]*)", r)
        # debian patches tachyon so it won't report the version
        # we hardcode '0.99' since that's indeed the version they ship
        return res[1] if res else '0.99'


tachyon_rt = TachyonRT()



from sage.structure.sage_object cimport SageObject
from sage.rings.padics.pow_computer cimport PowComputer_class
from sage.rings.padics.padic_generic_element cimport pAdicGenericElement

cdef class pAdicPrinter_class(SageObject):
    cdef object ring
    cdef int mode
    cdef bint pos
    cdef object ram_name
    cdef object latex_ram_name
    cdef object unram_name
    cdef object latex_unram_name
    cdef object var_name
    cdef object latex_var_name
    cdef object sep
    cdef object alphabet
    cdef PowComputer_class prime_pow
    cdef bint base
    cdef pAdicPrinter_class old
    cdef long max_ram_terms
    cdef long max_unram_terms
    cdef long max_terse_terms
    cdef object show_prec

    cdef base_p_list(self, value, bint pos)
    cdef _repr_gen(self, pAdicGenericElement elt, bint do_latex, bint pos, int mode, pname)
    cdef _repr_spec(self, pAdicGenericElement elt, bint do_latex, bint pos, int _mode, bint paren, pname)
    cdef _print_list_as_poly(self, L, bint do_latex, polyname, long expshift, bint increasing)
    cdef _truncate_list(self, L, max_terms, zero)
    cdef _var(self, x, exp, do_latex)
    cdef _dot_var(self, x, exp, do_latex)
    cdef _co_dot_var(self, co, x, exp, do_latex)
    cdef _plus_ellipsis(self, bint do_latex)
    cdef _ellipsis(self, bint do_latex)
    cdef _terse_frac(self, a, v, u, ram_name, bint do_latex)
    cdef _print_unram_term(self, L, bint do_latex, polyname, long max_unram_terms, long expshift, bint increasing)
    cdef _print_term_of_poly(self, s, coeff, bint do_latex, polyname, long exp)
